#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains scripts to demonstrate the features of the application.

Classes:
 - Examples: Contains examples and illustrations to demonstrate all request methods.
"""
from typing import Optional
import warnings

import os
import datetime
import pathlib

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.pyplot import GridSpec
from sqlalchemy import func, desc
from sqlalchemy.orm import Session

# noinspection PyUnresolvedReferences
import _paths  # pylint: disable=unused-import
from main import run as main_run
from model.database.tables import *
from model.utilities.export import database_session as get_session
from model.utilities.kill_switch import KillSwitch
from model.utilities.settings import Settings

warnings.filterwarnings("ignore")


class Examples:
    """
    Helper class providing examples and illustrations for all request methods.

    The respective configuration files are named according to the class-methods and can be found in the
    resources/configs folder. All requests are configured to terminate after a single run.
    """
    configuration_file: str
    plt.style.use("ggplot")
    pd.set_option("display.max_columns", 99)
    pd.set_option("expand_frame_repr", False)

    PATH = pathlib.Path.joinpath(_paths.all_paths.get("path_absolut"), "resources")

    @staticmethod
    def __start_catch_systemexit(configuration_file: str) -> None:
        try:
            main_run(configuration_file, os.getcwd())
        except SystemExit:
            return

    @staticmethod
    def __clear_database_table(session: Session, table: DatabaseTableType) -> None:
        """
        Deletes all entries from a database table.
        @param session: SQLAlchemy-ORM Session.
        @param table: Database table
        """
        print(f"Clearing table: {table.__name__}.")
        session.query(table).delete()
        session.commit()

    @staticmethod
    def __check_resources() -> Optional[bool]:
        """
        Check if the resources are available in the current working directory.
        """
        if not os.path.exists(Examples.PATH):
            print("Copy resources to your working directory first. "
                  "Use: runner.update_maps().")
            return False

    @staticmethod
    def static() -> plt.hist:
        """
        Request all available exchanges currency-pairs and create a histogram of their distribution.
        """
        if Examples.__check_resources() is False:
            return

        print("\nWarning: This example takes several minutes to complete. Do not interrupt the data requesting.")
        configuration_file = "examples/static"
        session = get_session(configuration_file)

        Examples.__start_catch_systemexit(configuration_file)

        query = session.query(ExchangeCurrencyPairView)
        dataframe = pd.read_sql(query.statement, con=session.bind)
        if dataframe.empty:
            return
        dataframe.exchange_name.value_counts().hist(bins=len(set(dataframe.exchange_name)))
        plt.title("Traded Pairs on Exchanges")
        plt.ylabel("Number of Exchanges")
        plt.xlabel("Number of Traded Pairs")
        plt.tight_layout()
        plt.show()

    @staticmethod
    def platforms() -> plt.plot:
        """
        Request BTC-USD data from the platform 'www.coingecko.com' and create a plot.
        """
        if Examples.__check_resources() is False:
            return

        configuration_file = "examples/platform"

        Examples.__start_catch_systemexit(configuration_file)

        session = get_session(configuration_file)
        query = session.query(HistoricRateView).filter(HistoricRateView.exchange == "COINGECKO",
                                                       HistoricRateView.first_currency == "BITCOIN",
                                                       HistoricRateView.second_currency == "USD")
        dataframe = pd.read_sql(query.statement, con=session.bind, index_col="time")
        if dataframe.empty:
            return
        dataframe.sort_index(inplace=True)

        fig = plt.figure(constrained_layout=True, figsize=(8, 6))
        grid_spec = GridSpec(4, 4, figure=fig)
        plt.rc("grid", linestyle=":", color="black")

        ax0 = fig.add_subplot(grid_spec[0:2, :])
        ax0.plot(dataframe.close, label="Close")
        plt.setp(ax0.get_xticklabels(), visible=False)
        plt.title("Bitcoin Daily Close in US-Dollar")
        ax0.grid(True)

        ax1 = fig.add_subplot(grid_spec[2:3, :])
        ax1.bar(dataframe.volume[dataframe.volume < 150 * 1e9].index,
                dataframe.volume[dataframe.volume < 150 * 1e9] / 1e9, label="Volume")
        plt.setp(ax1.get_xticklabels(), visible=False)
        ax1.grid(True)
        ax1.set_ylabel("Billion")
        plt.title("Bitcoin Daily Volume in US-Dollar")

        ax2 = fig.add_subplot(grid_spec[3:4, :])
        ax2.plot((dataframe.market_cap.divide(dataframe.close, axis=0) / 1e6), label="Supply")
        ax2.grid(True)
        ax2.set_ylabel("Million")
        ax2.set_xlabel("Time (Daily)")
        plt.title("Bitcoin Total Coin Supply")
        plt.tight_layout()
        plt.show()

    @staticmethod
    def minute_candles(timer: int = 60) -> plt.plot:
        """
        Request BTC-USD(T) data from several exchanges and plot them simultaneously.
        """
        if Examples.__check_resources() is False:
            return

        print(f"Note: The program will run for {timer} seconds before terminating automatically.")
        configuration_file = "examples/minute_candles"
        session = get_session(configuration_file)
        Examples.__clear_database_table(session, HistoricRate)

        switch: KillSwitch
        with KillSwitch() as switch:
            switch.set_timer(timer)
            Examples.__start_catch_systemexit(configuration_file)

        exchanges = ("BINANCE", "BITTREX", "HITBTC")
        session = get_session(configuration_file)
        query = session.query(HistoricRateView)
        query = query.filter(HistoricRateView.exchange.in_(exchanges))

        dataframe = pd.read_sql(query.statement, con=session.bind, index_col="time")
        if dataframe.empty:
            return
        dataframe = pd.pivot_table(dataframe, columns=dataframe.exchange, index=dataframe.index)
        dataframe = dataframe.close

        for column in dataframe.columns:
            plt.plot(dataframe.loc[:, column].dropna(), linestyle="dotted", linewidth=.75, label=column)
        plt.title("ETH/BTC - Minute Candles")
        plt.xlabel("Time")
        plt.ylabel("Price in US-Dollar")
        plt.legend()
        plt.xticks(rotation=45)
        plt.tight_layout()
        plt.show()

    @staticmethod
    def trades() -> Optional[pd.DataFrame]:
        """
        Request ETH-BTC transaction data from Coinbase and plot the price series and trade direction.
        """
        if Examples.__check_resources() is False:
            return

        configuration_file = "examples/trades"

        Examples.__start_catch_systemexit(configuration_file)

        exchange = "COINBASE"
        session = get_session(configuration_file)
        query = session.query(TradeView).filter(TradeView.exchange == exchange). \
            order_by(desc(TradeView.time)).limit(1000)

        dataframe = pd.read_sql(query.statement, con=session.bind, index_col="time")
        if dataframe.empty:
            return
        dataframe.sort_index(inplace=True)

        plt.plot(dataframe[dataframe.direction == "sell"].loc[:, "price"], linestyle="dotted",
                 color="red", label="Sells", linewidth=1.5)
        plt.plot(dataframe[dataframe.direction == "buy"].loc[:, "price"], linestyle="dotted",
                 color="green", label="Buys", linewidth=1.5)

        plt.xlabel("Timestamp")
        plt.xticks(rotation=45)
        plt.legend()
        plt.title("ETH/BTC Trades from Coinbase")

        plt.ylabel("Price in BTC")
        plt.tight_layout()
        plt.show()
        return dataframe.head(10)

    @staticmethod
    def order_books() -> Optional[pd.DataFrame]:
        """
        Requests the current order-book snapshot from Coinbase and plot the market depth.
        """
        if Examples.__check_resources() is False:
            return

        configuration_file = "examples/order_books"
        exchange = "COINBASE"
        session = get_session(configuration_file)
        Examples.__start_catch_systemexit(configuration_file)

        (timestamp,) = session.query(func.max(OrderBookView.time)).first()
        query = session.query(OrderBookView).filter(OrderBookView.exchange == exchange,
                                                    OrderBookView.time == timestamp,
                                                    OrderBookView.position <= 50)

        dataframe = pd.read_sql(query.statement, con=session.bind, index_col="time")
        if dataframe.empty:
            return

        # insert row that plot starts at amount = 0.
        timestamp = dataframe.index[0] - datetime.timedelta(days=1)
        new_data = dataframe.iloc[0, :].to_dict()
        template = {timestamp: {"bids_price": None, "bids_amount": 0, "asks_price": None,
                                "asks_amount": 0, "position": -1}}
        template.get(timestamp).update((k, new_data[k]) for k in set(new_data).intersection(template.get(timestamp))
                                       if template.get(timestamp).get(k) is None)

        dataframe = pd.concat([dataframe, pd.DataFrame.from_dict(template, orient="index")], axis=0)
        dataframe.sort_values(by="position", ascending=True, inplace=True)

        plt.step(dataframe.bids_price, dataframe.bids_amount.cumsum(), color="green", label="bids")
        plt.step(dataframe.asks_price, dataframe.asks_amount.cumsum(), color="red", label="asks")

        plt.ylim(ymin=0)
        plt.title("Market Depth BTC/USD(T)")
        plt.xlabel("Price in USD(T)")
        plt.ylabel("Accum. Size in BTC")
        plt.legend()
        plt.tight_layout()
        plt.show()
        return dataframe.iloc[1:10, :]

    @staticmethod
    def exchange_listings() -> plt.plot:
        """
        Collects historical data for 10 currency-pairs quoted against USD(T) and plots the amount of exchanges,
        each currency was listed on over time.
        """
        if Examples.__check_resources() is False:
            return

        print("\nWarning: This example takes several minutes to complete. Do not interrupt the data requesting.")
        configuration_file = "examples/exchange_listings"

        settings: Settings
        with Settings() as settings:
            settings.set("request_settings", "min_return_tuples", 100)
            settings.set("request_settings", "interval_settings", "equal")

            Examples.__start_catch_systemexit(configuration_file)

        session = get_session(configuration_file)
        base_currencies = ("BTC", "LINK", "ETH", "XRP", "LTC", "ATOM", "ADA", "XLM", "BCH", "DOGE")
        query = session.query(HistoricRateView.time,
                              HistoricRateView.exchange,
                              HistoricRateView.first_currency,
                              HistoricRateView.close).filter(HistoricRateView.first_currency.in_(base_currencies))
        dataframe = pd.read_sql(query.statement, con=session.bind, index_col="time")
        if dataframe.empty:
            return
        dataframe = pd.pivot_table(dataframe, columns=[dataframe.exchange, dataframe.first_currency],
                                   index=dataframe.index).close["2010-01-01":]

        for currency in base_currencies:
            temp = dataframe.loc[:, (slice(None), currency.upper())]
            temp = temp.resample("d").mean()
            temp = temp.resample("m").median()
            temp.count(axis=1).plot(label="/".join([currency, "USD(T)"]))

        plt.legend()
        plt.xlabel("Time (Monthly)")
        plt.ylabel("Number of Exchanges")
        plt.tight_layout()
        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore[no-untyped-def]
"""
Test module for testing exchange.Exchange class

Authors:
    Steffen Guenther

Since:
    07.07.2021

Version:
    07.07.2021
"""
import os
import oyaml as yaml
from model.exchange.exchange import Exchange, format_request_url
from model.database.tables import Exchange as ExchangeTable
from model.database.tables import Currency, ExchangeCurrencyPair, Ticker
import _paths  # pylint: disable=unused-import

path = os.getcwd() + "/open_crypto/tests/unit_tests"

with open(path + "/test_file.yaml", "r", encoding="UTF-8") as file:
    test_file: dict = yaml.load(file, Loader=yaml.FullLoader)


class TestExchange:
    """Test class for Exchange."""
    # pylint: disable=too-many-public-methods

    Exchange = Exchange(test_file, None, timeout=10, interval="seconds")

    def test_increase_interval(self) -> None:
        """
        Test function to increase interval from "seconds" to "minutes".
        """
        index = 2
        self.Exchange.interval = self.Exchange.interval_strings[index]
        self.Exchange.increase_interval()
        assert self.Exchange.interval == self.Exchange.interval_strings[index + 1]

    def test_increase_max_interval(self) -> None:
        """
        Test that function can not increase interval higher "days".
        """
        index = - 1
        self.Exchange.interval = self.Exchange.interval_strings[index]
        self.Exchange.increase_interval()
        assert self.Exchange.interval == self.Exchange.interval_strings[index]

    def test_decrease_interval(self) -> None:
        """
        Test function that decreases interval from "hours" to "minutes".
        """
        index = 2
        self.Exchange.interval = self.Exchange.interval_strings[index]
        self.Exchange.decrease_interval()
        assert self.Exchange.interval == self.Exchange.interval_strings[index - 1]

    def test_decrease_min_interval(self) -> None:
        """
        Test that function cannot decrease interval lower than "seconds".
        """
        index = 0
        self.Exchange.interval = self.Exchange.interval_strings[index]
        self.Exchange.decrease_interval()
        assert self.Exchange.interval == self.Exchange.interval_strings[index]

    def test_format_request_url_without_alias(self):
        """
        Test format request_url without alias. Currency-pair is supposed to be in directly in the url.
        """
        request_name = "tickers"
        currency1 = Currency(name="Test1")
        currency2 = Currency(name="Test2")
        exchange = ExchangeTable(name="TestExchange")

        currency_pairs = ExchangeCurrencyPair(first=currency1, second=currency2, exchange=exchange)

        request_urls = self.Exchange.extract_request_urls(test_file["requests"][request_name],
                                                          request_name=request_name,
                                                          request_table=Ticker,
                                                          currency_pairs=currency_pairs)[request_name]

        url, params = format_request_url(request_urls["url"], request_urls["pair_template"], 'test1-test2',
                                         currency_pairs, request_urls["params"])

        assert url == 'https://api.pro.coinbase.com/products/test1-test2/ticker'
        assert params == {}

    def test_format_request_url_with_alias(self):
        """
        Test format request_url with alias. Currency-pair is supposed to be in params.
        """
        request_name = "order_books"  #
        currency1 = Currency(name="Test1")
        currency2 = Currency(name="Test2")
        exchange = ExchangeTable(name="TestExchange")

        currency_pairs = ExchangeCurrencyPair(first=currency1, second=currency2, exchange=exchange)

        request_urls = self.Exchange.extract_request_urls(test_file["requests"][request_name],
                                                          request_name=request_name,
                                                          request_table=Ticker,
                                                          currency_pairs=currency_pairs)[request_name]

        url, params = format_request_url(request_urls["url"], request_urls["pair_template"], 'test1-test2',
                                         currency_pairs, request_urls["params"])

        assert url == 'https://api.pro.coinbase.com/products/book'
        assert params == {'level': 2, 'symbol': 'test1-test2'}

    #
    # def test_sort_order_book(self):
    #     """
    #     # ToDo
    #     """
    #     pass

    # def test_extract_request_urls(self):
    #     """
    #     # ToDo
    #     """
    #     pass
    #
    # def test_apply_currency_pair_format(self):
    #     """
    #     # ToDo
    #     """
    #     pass
    #
    # def test_format_currency_pairs(self):
    #     """
    #     # ToDo
    #     """
    #     pass
    #
    # def test_format_data(self):
    #     """
    #     # ToDo
    #     """
    #     pass
    #
    # def test_request(self):
    #     """
    #     # ToDo
    #     """
    #     pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Validator classes to validate the configuration file passed by the user. The module checks for existence, types and
values of necessary parameters. The file is executed in main.py before initializing the program.

Classes:
    - ConfigFileValidator(
        - ConfigYamlValidator(
            -> super().__init__(

            -    ConfigSectionValidator(value),
            -    ProgramSettingKeyValidator(value),
            -    ProgramSettingValueValidator(value)
            )
        )
    )
"""

from typing import Any, Text, Dict, Union

from pandas import Interval
from typeguard import check_type

# pylint: disable=too-many-lines
from model.validating.api_map_validators import LoadFileValidator, LoadYamlValidator
from model.validating.base import Report, Validator, CompositeValidator
from model.validating.errors import KeyNotInDictError, WrongTypeError, WrongValueError


# ToDo Debug
class ProgramConfigValidator(CompositeValidator):
    """Validator for an API Map file.

    Validator for validating a given configuration file.
    Consists of a FileLoadValidator, a YamlLoadValidator and an ConfigYamlValidator.

    Attributes:
        value:
            The file name or file path as Text.
    """

    value: Text

    def validate(self) -> bool:
        """Validates the value.

        Validates the value attribute while generating a validation Report.

        Returns:
            Whether further Validators may continue validating.
        """
        load_file = LoadFileValidator(self.value)
        is_file_loaded = load_file.validate()
        self.append_report(load_file)

        # if file did not load
        if not is_file_loaded:
            return False

        load_yaml = LoadYamlValidator(load_file)
        is_yaml_loaded = load_yaml.validate()
        self.append_report(load_yaml)

        # if yaml did not load
        if not is_yaml_loaded:
            return False

        config_file = ProgramConfigYamlValidator(load_yaml.get_result_value())
        can_continue = config_file.validate()
        try:
            for report in config_file.report.reports:
                self.append_report(report)
        except (TypeError, Exception):
            self.append_report(config_file)

        return can_continue

    def result(self) -> bool:
        """
        TODO: Fill out
        """
        if not self.report:
            self.validate()

        return True


class ProgramConfigYamlValidator(CompositeValidator):
    """Validator for loading Validators.

    Validator for calling the other required Validators.
    """

    def __init__(self, value: Dict[Text, Any]):
        """Constructor of ApiMapValidator.

        @param value:
            The value to get validated.
        """

        super().__init__(
            value,
            ConfigSectionValidator(value),
            ProgramSettingKeyValidator(value),
            ProgramSettingValueValidator(value)
        )


class ConfigSectionValidator(Validator):
    """
    Validates if all sections and blocks are present.

    """
    blocks = ["logging", "request_settings"]
    sections = ["dirname", "filename_format", "level", "exception_hook",
                "min_return_tuples", "interval_settings"]

    def validate(self) -> bool:

        try:

            for key in self.value:
                if key not in ConfigSectionValidator.blocks:
                    raise KeyNotInDictError(key, dict.fromkeys(ConfigSectionValidator.blocks))

            for block in ConfigSectionValidator.blocks:
                for key in self.value.get(block).keys():
                    if key not in ConfigSectionValidator.sections:
                        raise KeyNotInDictError(key, dict.fromkeys(ConfigSectionValidator.sections))

        except KeyNotInDictError as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report(f"Configuration contains all blocks: {ConfigSectionValidator.blocks} and"
                                 f" sections: {ConfigSectionValidator.sections}")
            return True


class ProgramSettingKeyValidator(Validator):
    """
    Validates if all necessary keys exist in the section 'operational_settings'.
    """

    sections = {"dirname": str,
                "filename_format": str,
                "level": str,
                "exception_hook": Union[int, bool],
                "min_return_tuples": int,
                "interval_settings": str,
                }

    def validate(self) -> bool:
        """
        Validates the value.
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        # Does key exist in config file
        try:
            merged_values = dict()
            for item in self.value.keys():
                merged_values.update(self.value.get(item))

            for key, val in ProgramSettingKeyValidator.sections.items():
                try:
                    check_type(key, merged_values.get(key), val)
                except TypeError as error:
                    raise WrongTypeError(val, type(self.value.get(key))) from error

        except (KeyNotInDictError, WrongTypeError) as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Program settings have valid types")
            return True


class ProgramSettingValueValidator(Validator):
    """
    Validates if all necessary keys have correctly specified values in the section 'operational_settings'.

    """
    sections = {"level": ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                "min_return_tuples": Interval(0, 1000000, "both"),
                "interval_settings": ["equal", "lower", "higher", "lower_or_equal", "high_or_equal"],
                }

    def validate(self) -> bool:
        """
        Validates the value.

        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        try:
            merged_values = dict()
            for item in self.value.keys():
                merged_values.update(self.value.get(item))

            for key, val in ProgramSettingValueValidator.sections.items():
                if not merged_values.get(key) in val:
                    raise WrongValueError(val, merged_values.get(key), key)

        except WrongValueError as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Program settings are valid")
            return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module providing several utility functions needed across the whole program.

Functions:
    - TYPE_CONVERSION: Dict providing functions for type conversions.
    - COMPARATOR: Dict providing function for comparison.
    - read_config: Function reading the config-file.
    - yaml_loader: Function for loading yaml-files.
    - load_program_config: Function loading the program configuration.
    - provide_ssl_context: Function providing an SSL-context if none is available.
    - get_exchange_names: Function returning all exchanges.
    - replace_list_item: Function replacing a specific item from a list.
    - get_all_exchanges_and_methods: Function returning all exchanges and supported request methods.
    - prepend_spaces_to_columns: Function prepending spaces to columns for readability.
    - handler: Exception handler logging uncaught exceptions.
    - signal_handler: Function recognizing kill signals and raising SystemExit.
    - init_logger: Function initializing the global logger.
    - install_packages: Installs missing but needed packages.
"""
import sys
import subprocess
import calendar
import datetime
import logging
import os
import ssl
from datetime import timedelta
from pathlib import Path
from typing import Any, Optional, Dict, List, Union

import certifi
import dateutil.parser
import oyaml as yaml
import pandas as pd

import _paths
from model.utilities.kill_switch import KillSwitch
from model.utilities.time_helper import TimeHelper, TimeUnit
from resources.configs.global_config import GlobalConfig

TYPE_CONVERSIONS = {
    ("float", "from_timestamp"): {
        "function": TimeHelper.from_timestamp,
        "params": 1
    },
    ("bool", "int"): {
        "function": int,
        "params": 0
    },
    ("float", "int"): {
        "function": int,
        "params": 0
    },
    ("int", "bool"): {
        "function": bool,
        "params": 0
    },
    ("int", "div"): {
        "function": lambda integer, div: integer / (1 * div),
        "params": 1
    },
    ("any", "value"): {
        "function": lambda number: float(number) > 0,
        "params": 0
    },
    ("str", "bool"): {
        "function": lambda string: string.lower() == "true",
        "params": 0
    },
    ("str", "int"): {
        "function": int,
        "params": 0
    },
    ("str", "float"): {
        "function": float,
        "params": 0
    },
    ("str", "float_absolut"): {
        "function": lambda string: abs(float(string)),
        "params": 0
    },
    ("str", "floatNA"): {
        "function": lambda string: float(string) if string != "N/A" else None,
        "params": 0
    },
    ("str", "strptime"): {
        "function": lambda string, *args: datetime.datetime.strptime(string, args[0]),
        "params": 1
    },
    ("strptime_w_f", "strptime_wo_f"): {
        "function": lambda string, *args: datetime.datetime.strptime(string.split(".")[0], *args),
        "params": 1
    },
    ("str", "split"): {
        "function": lambda string, *args: string.split(args[0])[args[1]] if args[0] in string else None,
        "params": 2
    },
    ("str", "splitupper"): {
        "function": lambda string, *args: string.split(args[0])[args[1]].upper(),
        "params": 2
    },
    ("str", "slice"): {
        "function": lambda string, *args: string[args[0]:args[1]],
        "params": 2
    },
    ("str", "upper"): {
        "function": lambda string: string.upper(),
        "params": 0
    },
    ("str", "lower"): {
        "function": lambda string: string.lower(),
        "params": 0
    },
    ("str", "dateparser"): {
        "function": dateutil.parser.parse,
        "params": 0
    },
    ("datetime", "strftime"): {
        "function": lambda time, *args: datetime.datetime.strftime(time, args[0]),
        "params": 1
    },
    ("dateparser", "totimestamp"): {
        "function": lambda time: int(time.timestamp()),
        "params": 0
    },
    ("datetime", "totimestamp"): {
        "function": lambda time: int(time.timestamp()),
        "params": 0
    },
    ("datetime", "totimestampms"): {
        "function": lambda time: int(round(time.timestamp() * 1000)),
        "params": 0
    },
    ("datetime", "utctotimestamp"): {
        "function": lambda time: calendar.timegm(time.utctimetuple()),
        "params": 0
    },
    ("strptime", "totimestamp"): {
        "function": lambda string, *args: int(datetime.datetime.timestamp(datetime.datetime.strptime(string, args[0]))),
        "params": 1
    },
    ("none", "nowstrptime"): {
        "function": lambda arg: TimeHelper.now().replace(hour=0, minute=0, second=0, microsecond=0),
        "params": 0
    },
    ("none", "now"): {
        "function": TimeHelper.now,
        "params": 0
    },
    ("none", "now_format"): {
        "function": lambda spec: format(TimeHelper.now(), spec),
        "params": 1
    },
    ("none", "constant"): {  # Returns the first argument
        "function": lambda *args: args[0],
        "params": 1
    },
    ("none", "range"): {
        "function": lambda: range(1),
        "params": 0
    },
    ("value", "map"): {
        # translate into buy/sell. Args: {0: 'buy', 1:'sell'} and arg[0] is the response value (i.e. 0/1)
        "function": lambda *args: {args[1]: args[2], args[3]: args[4]}[args[0]],
        "params": 4
    },
    ("str", "split_at_del_or_index"): {
        "function": lambda string, *args: string.split(args[0])[args[2]] if len(string) != len(
            string.split(args[0])[0]) else string[:args[1]] if args[2] == 0 else string[args[1]:],
        "params": 3  # delimiter, index, 0 or 1 aka. left or right
    },
    ("none", "now_timestamp"): {
        "function": lambda: int(TimeHelper.now_timestamp()),
        "params": 0
    },
    ("none", "now_timestampms"): {
        "function": lambda: int(TimeHelper.now_timestamp(TimeUnit.MILLISECONDS)),
        "params": 0
    },
    ("now", "timedelta"): {
        "function": lambda delta: int(TimeHelper.to_timestamp(TimeHelper.now() - timedelta(days=int(delta)))),
        "params": 1
    },
    ("datetime", "timedelta"): {
        "function": lambda time, interval, delta: int(
            TimeHelper.to_timestamp(time - timedelta(**{interval: int(delta)}))),
        "params": 2
    },
    ("utcfromtimestamp", "timedelta"): {
        "function": lambda time, interval, value: TimeHelper.from_timestamp(time) - timedelta(
            **{interval: value}) if
        isinstance(time, int) else dateutil.parser.parse(time) - timedelta(**{interval: value}),
        "params": 2
    },
    ("datetime", "timedeltams"): {
        "function": lambda time, interval, delta: int(
            TimeHelper.to_timestamp(time - timedelta(**{interval: int(delta)}))) * 1000,
        "params": 2
    },
    ("datetime", "timestamp"): {
        "function": lambda time: int(TimeHelper.to_timestamp(time)),
        "params": 0
    },
    ("datetime", "timestampms"): {
        "function": lambda time: int(TimeHelper.to_timestamp(time)) * 1000,
        "params": 0
    },
    ("datetime", "format"): {
        "function": format,
        "params": 1
    },
    ("timedelta", "from_timestamp"): {
        "function": lambda time, unit, spec: format(TimeHelper.from_timestamp(time, unit), spec),
        "params": 2
    },
    ("from_timestamp", "to_start"): {
        "function": lambda time, interval: TimeHelper.start_end_conversion(time, interval, False),
        "params": 1
    },
    ("from_timestamp", "to_end"): {
        "function": lambda time, interval: TimeHelper.start_end_conversion(time, interval, True),
        "params": 1
    }
}
"""
    Type Conversions used to convert extracted values from the API-Response into the desired type ("first", "second").
    The values are specified in the .yaml-file of each exchange under the "mapping" of each method.
    The function is called in the Mapping Class of utilities.py under the method convert_types().

    "first":
        The actual type extracted from the API-Request (.json)
    "second":
        The desired type to convert
    "function":
        the actual function to apply
    "params":
        the number of additional parameters needed
"""

COMPARATOR = {"equal": lambda x, y: x == y,
              "lower": lambda x, y: x < y,
              "lower_or_equal": lambda x, y: x <= y,
              "equal_or_lower": lambda x, y: x <= y,
              "higher": lambda x, y: x > y,
              "higher_or_equal": lambda x, y: x >= y,
              "equal_or_higher": lambda x, y: x >= y}
"""
Dict providing basic compare functionality.
"""


def read_config(file: Optional[str] = None,
                section: Optional[str] = None,
                reset: bool = False) -> Dict[str, Any]:
    """ Loads and reads in the configuration file specified by the user.

    @param file: Name of the config file.
    @type file: str
    @param section: Name of the section the information is stored in.
    @type section: str
    @param reset: Reset global config file to None.
    @type reset: bool

    @return: Parameters for the program as a dictionary.
             Keys are the names of the parameters in the config-file.
    @rtype: dict[str, Any]

    @raise KeyError: If the section does not exist in the config.
    """
    if reset and not file:
        GlobalConfig().set_file()

    if file:
        GlobalConfig().set_file(file)

    while True:
        try:
            filename = GlobalConfig().file
            config_yaml = open(filename, encoding="UTF-8")
            break
        except FileNotFoundError:
            try:
                available_files = os.listdir(os.path.dirname(filename))
                available_files = [file for file in available_files if not
                                   os.path.isdir(os.path.join(os.path.dirname(filename), file))]
                print(f"File not found. Retry! \nAvailable file(s): {', '.join(available_files)}")
            except FileNotFoundError:
                print("File not found. Retry!")
            finally:
                GlobalConfig().set_file()

    config_dict = yaml.load(config_yaml, Loader=yaml.FullLoader)
    config_yaml.close()

    if section is None:
        return config_dict

    for general_section in config_dict.keys():
        if section == general_section:
            return config_dict[general_section]

        for nested_section in config_dict[general_section].keys():
            if section == nested_section:
                return config_dict[general_section][nested_section]

    raise KeyError()


def yaml_loader(exchange: str, path: str = None) -> Dict[str, Any]:
    """ Loads, reads and returns the data of a .yaml-file specified by the param exchange.

    @param exchange: The file name to load (exchange).
    @type exchange: str

    @param path: path to the yaml-files
    @type path: str

    @return: Returns a dict of the loaded data from the .yaml-file.
    @rtype: dict

    @raise Exception: If the .yaml file could not be evaluated for a given exchange.
    """
    exchange = exchange.replace(" ", "")

    if not path:
        path = _paths.all_paths.get("yaml_path")

    path = _paths.all_paths.get("path_absolut").joinpath(Path(path))

    try:
        with open(path.joinpath(".".join([exchange, "yaml"])), "r", encoding="UTF-8") as file:
            return yaml.load(file, Loader=yaml.FullLoader)

    except FileNotFoundError as error:
        print(f"\nFile {path.joinpath('.'.join([exchange, 'yaml']))} not found.")
        logging.exception("Error loading yaml of %s.\n", exchange)
        raise SystemExit from error

    except Exception as error:
        print(f"\nError loading yaml of {exchange}. ")
        logging.exception("Error loading yaml of %s.\n", exchange)
        raise SystemExit from error


def load_program_config(return_path: bool = False) -> Union[str, Dict]:
    """ Loads the program configuration from a predefined directory. If the file is not found, it will return
    the template with default settings.

    @return: Program config.
    @rtype: dict
    """
    path = _paths.all_paths.get("program_config_path")

    if return_path:
        return path

    try:
        with open(path, "r", encoding="UTF-8") as file:
            return yaml.load(file, Loader=yaml.FullLoader)

    except FileNotFoundError:
        path = "resources/templates/program_config.yaml"
        with open(path, "r", encoding="UTF-8") as file:
            return yaml.load(file, Loader=yaml.FullLoader)


def get_exchange_names(yaml_path: str = None) -> Optional[List[str]]:
    """ Gives information about all exchange that the program will send
    requests to. This means if the name of a exchange is not part of the
    list that is returned, the program will not send any request to said
    exchange.

    @param: yaml-path
    @type yaml_path: str
    @return: Names from all the exchange, which have a .yaml-file in
             the directory described in YAML_PATH.
    @rtype: list[str]
    """
    if not yaml_path:
        yaml_path = _paths.all_paths.get("yaml_path")
    path_to_resources = Path.joinpath(_paths.all_paths.get("path_absolut"), Path(yaml_path))

    try:
        exchanges = os.listdir(path_to_resources)
        exchanges = [x.split(".yaml")[0] for x in exchanges if x.endswith(".yaml")]
        exchanges.sort()
    except FileNotFoundError:
        print(f"YAML files not found. The path {path_to_resources} is incorrect.")
        logging.error("Exchange YAML-files not found. Path %s seems incorrect.", path_to_resources)
        return

    return exchanges


def provide_ssl_context() -> ssl.SSLContext:
    """ Provides an SSL-Context if none is found beforehand. Especially UNIX machine with kernel "Darwin" may not
    provide an SSL-context for Python. To avoid connections without ssl-verification, this method returns a new
    default SSL-Context plugged into the request method.

    @return: SSLContext
    """
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
    ssl_context.verify_mode = ssl.CERT_REQUIRED
    ssl_context.check_hostname = True
    ssl_context.load_default_certs()

    # ToDo: Test under new MacOS!
    ssl_context.load_verify_locations(
        cafile=os.path.relpath(certifi.where()),
        capath=None,
        cadata=None)

    return ssl_context


def replace_list_item(replace_list: list, condition: str, value: str) -> list:
    """ Replaces a specific value from a list.

    @param replace_list: The list in which the value needs to be replaced
    @param condition: The value to be updated
    @param value: The new value
    @return: Updated list
    """
    for i, item in enumerate(replace_list):
        if item == condition:
            replace_list[i] = value
    return replace_list


def get_all_exchanges_and_methods() -> Dict[str, dict]:
    """ Returns the exchange names and all supported methods.

    @return: List of exchanges with supported request methods.
    @rtype: list
    """
    result_dict = dict()
    yaml_path = _paths.all_paths.get("yaml_path")

    exchanges = get_exchange_names(yaml_path=yaml_path)
    for exchange in exchanges:
        file = yaml_loader(exchange, path=yaml_path)
        result_dict.update({exchange: {method: True for method in list(file.get("requests").keys())}})

    return result_dict


def prepend_spaces_to_columns(dataframe: pd.DataFrame, space_count: int = 3) -> pd.DataFrame:
    """ Adds spaced between pd.DataFrame columns for easy readability.

    @param dataframe: Dataframe to append spaced to
    @type: pd.DataFrame
    @param space_count: Number of spaces
    @type: int
    @return: DataFrame with appended spaced.
    @rtype: pd.DataFrame
    """
    dataframe.replace(pd.NA, False, inplace=True)
    spaces = " " * space_count

    # ensure every column name has the leading spaces:
    if isinstance(dataframe.columns, pd.MultiIndex):
        for i in range(dataframe.columns.nlevels):
            level_new = [spaces + str(s) for s in dataframe.columns.levels[i]]
            dataframe.columns.set_levels(level_new, level=i, inplace=True)
    else:
        dataframe.columns = spaces + dataframe.columns

    # ensure every element has the leading spaces:
    dataframe = dataframe.astype(str)
    dataframe = spaces + dataframe
    return dataframe


def handler(ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
    """ Method to catch and log unexpected exceptions.

    @param ex_type: Exception type
    @param ex_value: Values causing the exception
    @param ex_traceback: Traceback attribute of the exception
    """
    logging.exception("Uncaught exception: %s: %s", ex_type, ex_value)


def signal_handler(signal_number: Any, stack: Any) -> None:
    """ Helper function to exit the program. When CTRL+C is hit,
    the program will shut down with exit code(0).
    """
    KillSwitch().reset()
    print("\nExiting program.")
    raise SystemExit


def init_logger(path: str, program_config: dict) -> None:
    # ToDO: Only create log-file if something is actually logged. Or delete empty log-files afterwards.
    """ Initializes the logger, specifies the path to the logging files,
    the logging massage as well as the logging level.

    @param path: Path to store the logging file. By default the CWD.
    @param program_config: Config file with some advanced program settings.
    """
    if not read_config(file=None, section="operation_settings").get('enable_logging', True):
        logging.disable()
    else:
        dirname = program_config["logging"].get("dirname", "resources/log/")
        os.makedirs(path + dirname, exist_ok=True)
        time_format = program_config["logging"].get("filename_format", "%Y-%m-%d_%H-%M-%S")
        file_name = path + dirname + f"{TimeHelper.now().strftime(time_format)}.log"

        logging.basicConfig(
            filename=file_name,
            level=program_config["logging"].get("level", "ERROR")
        )


def split_str_to_list(string: str, splitter: str = ",") -> List[str]:
    """ Splits a string into a list of string.

    @param string: A long string.
    @param splitter: The splitting parameter.
    @return: List of strings.
    """
    items = string.rsplit(splitter)

    # remove possible blanks from strings
    return [item.replace(" ", "") for item in items]


def install_packages(package_name: str) -> None:
    """ Installs missing packages.

    @param package_name: The name of the missing package.
    """
    permission: str = input(f"Module '{package_name}' not found. It will be installed. Proceed (Y / n)?")

    if permission.lower() in ["y", "j", "yes"]:
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])
            return
        except Exception as exc:
            raise ModuleNotFoundError from exc

    else:
        print("Module will not be installed. Either install it manually or change the database\n"
              "management system, e.g use 'Sqlite'.")
        raise SystemExit



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore[no-untyped-def]
"""
Contain test cases to test the persistence functionality.

Classes:
 - TestPersistResponse: Contains test cases to test the persistence functionality.
"""

from itertools import permutations

from model.database.db_handler import DatabaseHandler
from model.database.tables import metadata, ExchangeCurrencyPairView, Exchange, ExchangeCurrencyPair, Ticker, Currency
from model.utilities.time_helper import TimeHelper


class TestPersistResponse:
    """
    Test class for DatabaseHandler.
    """

    db_config = {
        "sqltype": "sqlite",
        "client": None,
        "user_name": None,
        "password": None,
        "host": None,
        "port": None,
        "db_name": None,
    }

    db_handler = DatabaseHandler(metadata, debug=True, **db_config)
    session = db_handler.session_factory()
    currencies = ["BTC", "ETH", "LTC", "XRP", "DIO", "DASH"]
    currency_pairs = permutations(currencies, 2)
    exchange_currency_pairs = [("TESTEXCHANGE",) + pair for pair in currency_pairs]
    db_handler.persist_exchange_currency_pairs(exchange_currency_pairs,
                                               is_exchange=True)

    def test_persist_exchange_currency_pairs(self):
        """
        TODO: Fill out
        """
        result = self.session.query(ExchangeCurrencyPairView).all()
        result = [(item.exchange_name, item.first_name, item.second_name) for item in result]

        assert self.exchange_currency_pairs == result

    def test_persist_valid_ticker(self):
        """
        TODO: Fill out
        """
        response = [
            (TimeHelper.now(), TimeHelper.now(), 1, 1.0, 1.0, 1.0),
            (TimeHelper.now(), TimeHelper.now(), 2, 2.0, 2.0, 2.0),
            (TimeHelper.now(), TimeHelper.now(), 3, 3.0, 3.0, 3.0),
            (TimeHelper.now(), TimeHelper.now(), 4, 4.0, 4.0, 4.0)]

        exchanges_with_pairs = {
            self.session.query(Exchange).first(): dict.fromkeys(list(self.session.query(ExchangeCurrencyPair).limit(4)))
        }

        exchange = list(exchanges_with_pairs.keys())[0]
        mappings = ["start_time", "time", "best_ask", "best_bid", "last_price", "exchange_pair_id"]

        formatted_response = [(response, mappings)]
        self.db_handler.persist_response(exchanges_with_pairs,
                                         exchange,
                                         Ticker,
                                         iter(formatted_response))
        result = self.session.query(Ticker).all()

        result = [(item.exchange_pair.exchange.name,
                   item.exchange_pair.first.name,
                   item.exchange_pair.second.name,
                   item.start_time,
                   item.time,
                   item.exchange_pair_id,
                   item.last_price,
                   item.best_ask,
                   item.best_bid) for item in result]

        response2 = [(ExCuPair[0], ExCuPair[1], ExCuPair[2]) + res for (ExCuPair, res) in
                     zip(self.exchange_currency_pairs, response)]
        assert response2 == result
        self.session.query(Ticker).delete()

    def test_persist_response_with_unknown_column(self):
        """
        TODO: Fill out
        """
        response = [
            (TimeHelper.now(), TimeHelper.now(), 1.0, 1.0, 1.0, 1, None),
            (TimeHelper.now(), TimeHelper.now(), 2.0, 2.0, 2.0, 2, None),
            (TimeHelper.now(), TimeHelper.now(), 3.0, 3.0, 3.0, 3, None),
            (TimeHelper.now(), TimeHelper.now(), 4.0, 4.0, 4.0, 4, None)]

        exchanges_with_pairs = {
            self.session.query(Exchange).first(): dict.fromkeys(list(self.session.query(ExchangeCurrencyPair).limit(4)))
        }

        exchange = list(exchanges_with_pairs.keys())[0]
        mappings = ["start_time", "time", "best_ask", "best_bid", "last_price", "exchange_pair_id", "some_column"]

        formatted_response = [(response, mappings)]
        self.db_handler.persist_response(exchanges_with_pairs,
                                         exchange,
                                         Ticker,
                                         iter(formatted_response))
        result = self.session.query(Ticker).all()

        result = [(item.exchange_pair.exchange.name,
                   item.exchange_pair.first.name,
                   item.exchange_pair.second.name,
                   item.start_time,
                   item.time,
                   item.best_ask,
                   item.best_bid,
                   item.last_price,
                   item.exchange_pair_id,) for item in result]

        response2 = [(ExCuPair[0], ExCuPair[1], ExCuPair[2]) + res[:-1] for (ExCuPair, res) in
                     zip(self.exchange_currency_pairs, response)]
        assert response2 == result
        self.session.query(Ticker).delete()

    def test_persist_response_with_unknown_currency_pair(self):
        """
        TODO: Fill out
        """
        response = [("TEST1", "TEST2", TimeHelper.now(), TimeHelper.now(), 1.0, 1.0, 1.0, 1.0, 1)]
        exchanges_with_pairs = {self.session.query(Exchange).first(): []}

        exchange = list(exchanges_with_pairs.keys())[0]
        mappings = ["currency_pair_first", "currency_pair_second", "start_time", "time", "best_ask", "best_bid",
                    "last_price", "daily_volume"]

        formatted_response = [(response, mappings)]
        self.db_handler.persist_response(exchanges_with_pairs,
                                         exchange,
                                         Ticker,
                                         iter(formatted_response))

        result = self.session.query(ExchangeCurrencyPairView).filter(ExchangeCurrencyPairView.first_name == "TEST1",
                                                                     ExchangeCurrencyPairView.second_name == "TEST2").first()
        assert ("TESTEXCHANGE", "TEST1", "TEST2") == (result.exchange_name,
                                                      result.first_name,
                                                      result.second_name)
        self.session.query(Ticker).delete()

    def test_persist_response_with_none(self):
        """
        TODO: Fill out
        """
        response = [
            (TimeHelper.now(), TimeHelper.now(), None, 1.0, 1.0, 1),
            (TimeHelper.now(), TimeHelper.now(), 2.0, None, 2.0, 2),
            (TimeHelper.now(), TimeHelper.now(), 3.0, 3.0, None, 3),
            (TimeHelper.now(), TimeHelper.now(), 4.0, 4.0, 4.0, 4), ]

        exchanges_with_pairs = {self.session.query(Exchange).first():
                                    dict.fromkeys(list(self.session.query(ExchangeCurrencyPair).limit(4)))}

        exchange = list(exchanges_with_pairs.keys())[0]
        mappings = ["start_time", "time", "best_ask", "best_bid", "last_price", "exchange_pair_id"]

        formatted_response = [(response, mappings)]
        self.db_handler.persist_response(exchanges_with_pairs,
                                         exchange,
                                         Ticker,
                                         iter(formatted_response))

        result = self.session.query(Ticker).all()
        result = [(item.exchange_pair.exchange.name,
                   item.exchange_pair.first.name,
                   item.exchange_pair.second.name,
                   item.start_time,
                   item.time,
                   item.best_ask,
                   item.best_bid,
                   item.last_price,
                   item.exchange_pair_id,) for item in result]

        response2 = [(ExCuPair[0], ExCuPair[1], ExCuPair[2]) + res for (ExCuPair, res) in
                     zip(self.exchange_currency_pairs, response)]
        assert response2 == result
        self.session.query(Ticker).delete()

    def test_get_all_currency_pairs_from_exchange_with_no_invalid_pair(self):
        """
        Test for the method get_all_currency_pairs_from_exchange. This method will be called with the test dataset. The
        list of the id's of the exchange, first and second currency will be compared. All currency pairs from the
        test dataset should be returned of the method call, because all pairs are from the given exchange 'TESTEXCHANGE'.
        """
        test_result = self.db_handler.get_all_currency_pairs_from_exchange("TESTEXCHANGE")
        test_result = [(item.exchange_id,
                        item.first_id,
                        item.second_id) for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).all()
        result = [(item.exchange_id,
                   item.first_id,
                   item.second_id) for item in result]
        assert result == test_result

    def test_get_all_currency_pairs_from_exchange_with_invalid_pair(self):
        """
        Test for the method get_all_currency_pairs_from_exchange. This method will be called with the test dataset. An
        empty list of currency pairs should be returned, because in the given dataset of this testmethod are no
        currency pairs with the given exchange 'TESTEXCHANGE'.
        """
        self.session.query(ExchangeCurrencyPair).delete()
        self.db_handler._persist_exchange_currency_pair("invalid", "BTC", "ETH", True)  # pylint: protected-access
        test_result = self.db_handler.get_all_currency_pairs_from_exchange("TESTEXCHANGE")
        result = []
        assert result == test_result

        self.session.query(ExchangeCurrencyPair).delete()
        self.db_handler.persist_exchange_currency_pairs(self.exchange_currency_pairs,
                                                        is_exchange=True)

    def test_get_currency_pairs_with_first_currency_valid_1(self):
        """
        Test for the method get_currency_with_first_currency. This method will be called with the test dataset and 'BTC'
        as a currency. The list of the id's of the first currency will be compared.
        """
        test_result = self.db_handler.get_currency_pairs_with_first_currency("TESTEXCHANGE", ["BTC"])
        test_result = [item.first_id for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.first_id.__eq__(1)).all()
        result = [item.first_id for item in result]
        assert result == test_result

    def test_get_currency_pairs_with_first_currency_valid_2(self):
        """
        Test for the method get_currency_with_first_currency. This method will be called with the testd ataset and 'BTC'
        and 'LTC' as a currency. The list of the id's of the first currency will be compared.
        """
        test_result = self.db_handler.get_currency_pairs_with_first_currency("TESTEXCHANGE", ["BTC", "LTC"])
        test_result = [item.first_id for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.first_id.__eq__(1)).all()
        result.extend(self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.first_id.__eq__(3)).all())
        result = [item.first_id for item in result]
        assert result == test_result

    def test_get_currency_pairs_with_first_currency_invalid(self):
        """
        Test for the method get_currency_with_first_currency. This method will be called with the test dataset and 'BTC'
        as a currency. The list of the id's of the first currency will be compared. An Empty list should be returned,
        because there are no currency pairs with the currency 'BAT'.
        """
        test_result = self.db_handler.get_currency_pairs_with_first_currency("TESTEXCHANGE", ["BAT"])
        result = []
        assert result == test_result

    def test_get_currency_pairs_with_second_currency_valid(self):
        """
        Test for the method get_currency_with_second_currency. This method will be called with the test dataset and 'BTC'
        as a currency. The List of the id's of the second currency will be compared.
        """
        # todo : Eingabeparameter in der Methode get_currency_pairs_with_second_currency in db_handler
        #       müsste eigentlich eine Liste an Currencies entgegennehmen, wie in der Methode
        #       get_currency_pairs_with_first_currency, und nicht nur einen einzelnen String.
        # Ich habe das noch nicht gefixt, da ich nicht genau weiß, ob dann eventuell Fehlermeldungen geworfen werden,
        # bei den vorhandenen Aufrufen der Methode. Diese Aufrufe müssten dann eventuell angepasst werden.
        test_result = self.db_handler.get_currency_pairs_with_second_currency("TESTEXCHANGE", ["BTC"])
        test_result = [item.second_id for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.second_id.__eq__(1)).all()
        result = [item.second_id for item in result]
        assert result == test_result

    def test_get_currency_pairs_with_second_currency_invalid(self):
        """
        Test for the method get_currency_with_second_currency. This method will be called with the test dataset and 'BTC'
        as a currency. The list of the id's of the second currency will be compared. An Empty list should be returned,
        because there are no currency pairs with the currency 'BAT'.
        """
        test_result = self.db_handler.get_currency_pairs_with_second_currency("TESTEXCHANGE", ["BAT"])
        result = []
        assert result == test_result

    def test_persist_exchange_and_get_exchange_id(self):
        """
        Test for the methods persist_exchange and get_exchange_id. The method persist_exchange will be called to persist
        a new test exchange. The return of the method get_exchange_id will be compared. Afterwards the new test exchange
        will be deleted from the test dataset.
        """
        self.db_handler.persist_exchange("TEST", True)
        test_result = self.db_handler.get_exchange_id("TEST")
        result = self.session.query(Exchange).all()
        for item in result:
            if item.name == "TEST":
                result_id = item.id

        assert result_id == test_result

        self.session.query(Exchange).filter(Exchange.id.__eq__(result_id)).delete()

    def test_get_currency_id(self):
        """
        Test for the method get_currency_id. This method will be called. The returned id's will be compared.
        """
        test_result = self.db_handler.get_currency_id("BTC")
        result = self.session.query(Currency).all()
        for item in result:
            if item.name == "BTC":
                result_id = item.id
        assert result_id == test_result

    def test_get_currency_pairs(self):
        """
        Test for the method get_currency_pairs. This method will be called with a given test exchange and a given list of
        dictionaries (representing currency pairs).
        For simplicity, only the id's will be compared (not the whole objects).
        """
        currency_pairs = [{"first": "BTC", "second": "LTC"},
                          {"first": "BTC", "second": "DIO"}]
        test_result = self.db_handler.get_currency_pairs("TESTEXCHANGE", currency_pairs)
        test_result = [(item.exchange_id,
                        item.first_id,
                        item.second_id) for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.exchange_id.__eq__(1),
                                                                 ExchangeCurrencyPair.first_id.__eq__(1),
                                                                 ExchangeCurrencyPair.second_id.__eq__(3)).all()
        result.extend(self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.exchange_id.__eq__(1),
                                                                      ExchangeCurrencyPair.first_id.__eq__(1),
                                                                      ExchangeCurrencyPair.second_id.__eq__(5)).all())
        result = [(item.exchange_id,
                   item.first_id,
                   item.second_id) for item in result]
        assert result == test_result

    def test_get_exchange_currency_pairs1(self):
        """
        Test for the method get_exchange_currency_pairs. This method will be called with a given test exchange, list of
        dictionaries (representing currency pairs), al ist of first currencies and a list of second currencies.
        For simplicity, only the id's will be compared (not the whole objects).
        """
        # currency_pairs = [{"first": "BTC", "second": "LTC"}]
        currency_pairs = "BTC-LTC"
        firsts = "DIO"
        seconds = None
        test_result = self.db_handler.get_exchanges_currency_pairs("TESTEXCHANGE", currency_pairs, firsts, seconds)
        test_result = [(item.exchange_id,
                        item.first_id,
                        item.second_id) for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.exchange_id.__eq__(1),
                                                                 ExchangeCurrencyPair.first_id.__eq__(1),
                                                                 ExchangeCurrencyPair.second_id.__eq__(3)).all()
        result.extend(self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.exchange_id.__eq__(1),
                                                                      ExchangeCurrencyPair.first_id.__eq__(5)).all())
        result = [(item.exchange_id,
                   item.first_id,
                   item.second_id) for item in result]
        assert result == test_result

    def test_get_exchange_currency_pairs2(self):
        """
        Test for the method get_exchange_currency_pairs. This method will be called with a given test exchange, list of
        dictionaries (representing currency pairs), al ist of first currencies and a list of second currencies.
        For simplicity, only the id's will be compared (not the whole objects).
        """
        currency_pairs = "BTC-DASH"
        firsts = "XRP"
        seconds = "ETH"
        test_result = self.db_handler.get_exchanges_currency_pairs("TESTEXCHANGE", currency_pairs, firsts, seconds)
        test_result = [(item.exchange_id,
                        item.first_id,
                        item.second_id) for item in test_result]
        result = self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.exchange_id.__eq__(1),
                                                                 ExchangeCurrencyPair.first_id.__eq__(1),
                                                                 ExchangeCurrencyPair.second_id.__eq__(6)).all()
        result.extend(self.session.query(ExchangeCurrencyPair).filter(ExchangeCurrencyPair.exchange_id.__eq__(1),
                                                                      ExchangeCurrencyPair.first_id.__eq__(4),
                                                                      ExchangeCurrencyPair.second_id.__eq__(2)).all())

        result = [(item.exchange_id,
                   item.first_id,
                   item.second_id) for item in result]
        result = list(dict.fromkeys(result))  # remove duplicates from list
        assert result == test_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Validators to validate the CSV export template.

The module checks for existence, types and values of necessary parameters.

Validators structure:

- CsvExportTemplateValidator
  - LoadFileValidator
  - LoadYamlValidator
  - ExportTemplateValidator
    - BlockValidator
    - BlockExportValidator
    - BlockDatabaseValidator
    - BlockQueryOptionsValidator
"""
import sys
from typing import Any, Text, Dict

from model.validating.api_map_validators import LoadFileValidator, LoadYamlValidator
from model.validating.base import Report, Validator, CompositeValidator
from model.validating.errors import KeyNotInDictError, WrongTypeError, WrongValueError
from model.validating.errors import WrongCurrencyPairFormatError


class CsvExportTemplateValidator(CompositeValidator):
    """
    Validator for an CSV export template.

    Attributes:
        - value: The file name or file path as Text.
    """
    value: Text

    def validate(self) -> bool:
        """
        Validates the value.

        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        load_file = LoadFileValidator(self.value)
        is_file_loaded = load_file.validate()
        self.append_report(load_file)

        if not is_file_loaded:
            return False

        load_yaml = LoadYamlValidator(load_file)
        is_yaml_loaded = load_yaml.validate()
        self.append_report(load_yaml)

        if not is_yaml_loaded:
            return False

        export_template = ExportTemplateValidator(load_yaml.get_result_value())
        can_continue = export_template.validate()
        try:
            for report in export_template.report.reports:
                self.append_report(report)
        except (TypeError, Exception):
            self.append_report(export_template)

        return can_continue

    def result(self) -> bool:
        """
        TODO: Fill out
        """
        if not self.report:
            self.validate()

        return True


class ExportTemplateValidator(CompositeValidator):
    """
    Validator for the contents of an CSV export template.
    """

    def __init__(self, value: Dict[Text, Any]):
        """
        Constructor of ExportTemplateValidator.

        @param value: The value to get validated.
        """
        super().__init__(
            value,
            BlockValidator(value),
            BlockExportValidator(value["export"]),
            BlockDatabaseValidator(value["database"]),
            BlockQueryOptionsValidator(value["query_options"]),
        )


class BlockValidator(Validator):
    """
    Validator to check if all sections and blocks are present.
    """
    block = ["export", "database", "query_options"]

    def validate(self) -> bool:
        try:
            for key in self.value:
                if key not in BlockValidator.block:
                    raise KeyNotInDictError(key, self.value)
        except KeyNotInDictError as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report(f"Export template contains all blocks: {BlockValidator.block}")
            return True


class BlockExportValidator(Validator):
    """
    Validator to check if the export block is valid.
    """

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "delimiter" not in self.value:
                raise KeyNotInDictError("delimiter", self.value)

            if not isinstance(self.value["delimiter"], str):
                raise WrongTypeError(str, type(self.value["delimiter"]))

            if "decimal" not in self.value:
                raise KeyNotInDictError("decimal", self.value)

            if not isinstance(self.value["decimal"], str):
                raise WrongTypeError(str, type(self.value["decimal"]))

            if self.value["decimal"] not in [".", ",", " "]:
                raise WrongValueError([".", ",", " "], self.value["decimal"], "decimal")

        except (KeyNotInDictError, WrongTypeError, WrongValueError) as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report("Export block is valid")
            return True


class BlockDatabaseValidator(Validator):
    """
    Validator to check if the database block is valid.
    """
    db_strings = {
        "sqlite": ["sqltype", "db_name"],
        "mariadb": ["sqltype", "user_name", "password", "host", "port", "db_name"],
        "mysql": ["sqltype", "user_name", "password", "host", "port", "db_name"],
        "postgres": ["sqltype", "user_name", "password", "host", "port", "db_name"],
    }

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "sqltype" not in self.value:
                raise KeyNotInDictError("sqltype", self.value)

            for key in BlockDatabaseValidator.db_strings.get(self.value.get("sqltype")):

                if key not in self.value:
                    raise KeyNotInDictError(key, self.value)

                if self.value.get(key) is None:
                    raise WrongTypeError(str, type(self.value.get(key)), key)

        except (KeyNotInDictError, WrongTypeError) as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report("Database connection string is valid")
            return True


class BlockQueryOptionsValidator(Validator):
    """
    Validator to check if the query options block is valid.
    """

    def validate(self) -> bool:
        """
        Validates the value.

        @return: Whether further Validators may continue validating.
        """
        try:
            if "table_name" not in self.value:
                raise KeyNotInDictError("table_name", self.value)

            if not isinstance(self.value["table_name"], str):
                raise WrongTypeError(str, type(self.value["table_name"]))

            if self.value["table_name"] not in ["HistoricRate", "Ticker", "Trade", "OrderBook"]:
                raise WrongValueError(
                    ["HistoricRate", "Ticker", "Trade", "OrderBook"], self.value["table_name"], "table_name"
                )

            if "query_everything" not in self.value:
                raise KeyNotInDictError("query_everything", self.value)

            if not isinstance(self.value["query_everything"], bool):
                raise WrongTypeError(bool, type(self.value["query_everything"]))

            if "from_timestamp" not in self.value:
                raise KeyNotInDictError("from_timestamp", self.value)

            if not isinstance(self.value["from_timestamp"], (str, type(None))):
                raise WrongTypeError((str, type(None)), type(self.value["from_timestamp"]))

            if "to_timestamp" not in self.value:
                raise KeyNotInDictError("to_timestamp", self.value)

            if not isinstance(self.value["to_timestamp"], (str, type(None))):
                raise WrongTypeError((str, type(None)), type(self.value["to_timestamp"]))

            if "exchanges" not in self.value:
                raise KeyNotInDictError("exchanges", self.value)

            if not isinstance(self.value["exchanges"], (str, type(None))):
                raise WrongTypeError((str, type(None)), type(self.value["exchanges"]))

            if "currency_pairs" not in self.value:
                raise KeyNotInDictError("currency_pairs", self.value)

            if not isinstance(self.value["currency_pairs"], (str, type(None))):
                raise WrongTypeError((str, type(None)), type(self.value["currency_pairs"]))

            if pair_string := self.value["currency_pairs"]:
                if pair_string == "all" or pair_string.count("-") >= 0 and pair_string.count("-") == pair_string.count(
                        ",") + 1:
                    raise WrongCurrencyPairFormatError(["-", ","], pair_string, "currency_pairs")

            if "first_currencies" not in self.value:
                raise KeyNotInDictError("first_currencies", self.value)

            if not isinstance(self.value["first_currencies"], (str, type(None))):
                raise WrongTypeError((str, type(None)), type(self.value["first_currencies"]))

            if "second_currencies" not in self.value:
                raise KeyNotInDictError("second_currencies", self.value)

            if not isinstance(self.value["second_currencies"], (str, type(None))):
                raise WrongTypeError((str, type(None)), type(self.value["second_currencies"]))

        except (KeyNotInDictError, WrongTypeError, WrongValueError) as error:
            self.report = Report(error)
            return False
        else:
            self.report = Report("Export block is valid")
            return True


if __name__ == "__main__":
    IS_VALID = CsvExportTemplateValidator("open_crypto/resources/templates/csv_export_template.yaml").validate()
    sys.exit(int(not IS_VALID))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
TODO: Fill out module docstring.
"""

import asyncio
import json
import os
from typing import List

from model.database.db_handler import DatabaseHandler
from model.database.tables import metadata
from model.exchange.exchange import Exchange
from model.utilities.utilities import read_config, get_exchange_names, yaml_loader

job_name: str = "historic_rates"  # currency-pairs, historic-rates...
yaml_path: str = "../resources/running_exchanges/all/"
saving_path: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), "json_test_responses")

db_params = read_config("database")
database_handler = DatabaseHandler(metadata, **db_params)

exchange_names = get_exchange_names()

print(saving_path)
exchanges: List[Exchange] = list()
for exchange_name in exchange_names:
    exchanges.append(Exchange(yaml_loader(exchange_name), None, None))

for exchange in exchanges:
    # Making sure every exchange has all the available currency-pairs
    # pair_response = asyncio.run(exchange.request_currency_pairs('currency_pairs'))
    # pairs = exchange.format_currency_pairs(pair_response)
    # database_handler.persist_exchange_currency_pairs(pairs, True)

    all_currency_pairs_for_ex = database_handler.get_all_currency_pairs_from_exchange(exchange.name)
    responses = asyncio.run(exchange.request(job_name, all_currency_pairs_for_ex))[2]
    for currency_pair in responses.keys():
        full_saving_path = os.path.join(saving_path, exchange.name, job_name)
        if not os.path.exists(full_saving_path):
            os.makedirs(full_saving_path)
        if currency_pair:
            first_currency: str = currency_pair.first.name
            second_currency: str = currency_pair.second.name
            json_file = open(os.path.join(full_saving_path,
                                          f"{exchange.name.upper()}_{first_currency}_{second_currency}"),
                             "w", encoding="UTF-8")
        else:
            json_file = open(os.path.join(full_saving_path, f"{exchange.name.upper()}_ALL"), "w", encoding="UTF-8")

        json.dump(responses[currency_pair], json_file, indent=4)
        json_file.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Contains exceptions to represent validation failures.
"""

from typing import Text, Any, Iterable, Union, Type, Dict, List

import validators


class ValidationError(Exception):
    """
    Exception in case a validated value is not valid.
    """


class KeyNotInDictError(ValidationError):
    """
    Exception in case a required key is not included in a particular dict.

    Attributes:
        missing_key:
            A key that is not included in a dict.
        inspected_dict:
            A dict in which a certain key should be included.
    """

    def __init__(self, missing_key: Text, inspected_dict: Dict[Text, Any]):
        """
        Constructor of KeyNotInDictError.

        @param missing_key: A key that is not included in a dict.
        @param inspected_dict: A dict in which a certain key should be included.
        """
        super().__init__("Key was not in Dict.")
        self.missing_key = missing_key
        self.inspected_dict = inspected_dict

    def __str__(self) -> Text:
        """
        A method for representing a text.

        A text value returning the missing key, which is no part of the
        inspected dict.

        @return: A Text.
        """
        return f"Key '{self.missing_key}' not in keys {list(self.inspected_dict.keys())}"


class KeyNotIntendedError(ValidationError):
    """
    Exception in case that a key is not intended to be in a certain dict.

    Attributes:
        intended_keys:
            A set of keys, which are intended to be a part of a dict.
        actual_key:
            A key, which is not intended to be in a dict.
    """

    def __init__(self, intended_keys: Iterable[Text], actual_key: Text):
        """
        Constructor of KeyNotIntendedError.

        @param intended_keys: A set of keys, which are intended to be a part of a dict.
        @param actual_key: A key, which is not intended to be in a dict.
        """
        super().__init__("Key was not intended to be in Dict.")
        self.intended_keys = set(intended_keys)
        self.actual_key = actual_key

    def __str__(self) -> Text:
        """
        A method for representing a text.

        A text value returning the actual key, which is not intended to be part
        of a certain dict and the intended keys, which are allowed to be in the
        dict.

        @return: A Text.
        """
        return f"Key '{self.actual_key}' not intended to be in Dict, Allowed: {self.intended_keys}."


class SubstringNotInStringError(ValidationError):
    """
    Exception in case that a substring is not part of a certain string.

    Attributes:
        missing_substring:
            A substring, which is no a part of a string.
        inspected_string:
            A string, in which the substring shall be contained.
    """

    def __init__(self, missing_substring: Text, inspected_string: Text):
        """
        Constructor of SubstringNotInStringError.

        @param missing_substring: A substring, which is not a part of a string.
        @param inspected_string: A string, in which the substring shall be contained.
        """
        super().__init__("Expected Substring was not found in String.")
        self.missing_substring = missing_substring
        self.inspected_string = inspected_string

    def __str__(self) -> Text:
        """
        A method for representing a text.

        A text value returning the missing substring, which is not a part of
        the inspected string.

        @return: A Text.
        """
        return f"Substring '{self.missing_substring}' not in '{self.inspected_string}'"


class WrongTypeError(ValidationError):
    """
    Exception in case that a value has a wrong type.

    Attributes:
        expected_type:
            A type that is expected.
        actual_type:
            The actual type, which is not the expected type.
    """

    def __init__(
            self,
            expected_type: Union[Type, Iterable[Type]],
            actual_type: Type,
            key: str = None):
        """
        Constructor of WrongTypeError.

        @param expected_type: A type that is expected.
        @param actual_type: The actual type, which is not the expected type.
        """
        super().__init__("Value has wrong type.")
        self.expected_type = set(expected_type) if isinstance(expected_type, Iterable) else expected_type
        self.actual_type = actual_type
        self.key = key

    def __str__(self) -> Text:
        """
        A method for representing a text.

        A text value returning the expected type(s), which is/are unlike the
        actual type.

        @return: A Text.
        """
        return f"{'Key ' + self.key + ': ' if self.key else ''}Expected type(s) '{self.expected_type}' " \
               f"!= actual type '{self.actual_type}'."


class UrlValidationError(ValidationError):
    """
    Exception in case that a URL is not valid.
    """

    def __init__(self, url: Text, report: validators.ValidationFailure = None):
        """
        Constructor of UrlValidationError.

        @param url: A URL that shall be checked.
        @param report: The report when a URL is not valid.
        """
        super().__init__("URL was not valid.")
        self.url = url
        self.report = report

    def __str__(self) -> Text:
        """
        A method for representing a text.

        A text value returning the invalid URL.

        @return: A Text.
        """
        return f"URL '{self.url}' is not valid."


class NamingConventionError(ValidationError):
    """
    Exception in case that the naming convention is violated.
    """

    def __init__(self, naming_pattern: Text, name: Text):
        """
        Constructor of NamingConventionError.

        @param naming_pattern: A given naming pattern.
        @param name: The name which shall be checked.
        """
        super().__init__("Naming convention was violated.")
        self.naming_pattern = naming_pattern
        self.name = name

    def __str__(self) -> Text:
        """
        A method for representing a text.

        A text value returning the name, which does not match the naming
        convention.

        @return: A Text.
        """
        return f"'{self.name}' did not match naming convention's Regex Pattern '{self.naming_pattern}'."


class WrongValueError(ValidationError):
    """
    Exception in case that a value has a wrong type.
    """

    def __init__(
            self,
            expected_value: List[Any],
            actual_value: Union[str, int, float],
            key: str):
        """
        Constructor of WrongValueError.

        Args:
            expected_value:
                A value that is expected.
            actual_value:
                The actual value, which is not the expected value.
        """
        super().__init__("Key has wrong value.")
        self.expected_value = expected_value \
            if not isinstance(expected_value, Iterable) else set(expected_value)
        self.actual_value = actual_value
        self.key = key

    def __str__(self) -> Text:
        """A method for representing a text.

        A text value returning the expected value(s), which is/are unlike the
        actual value.

        @return: A Text.
        """
        return f"Expected value(s) '{self.expected_value}' != actual value '{self.actual_value}' in '{self.key}'."


class WrongCompositeValueError(ValidationError):
    """
    Exception in case that a composition of values have the wrong type.

    Attributes:
        keys:
            The keys containing the wrong values.

    """

    def __init__(self, keys: List[Union[str, Type]]):
        """
        Constructor of WrongCompositeValueError.

        Args:
            keys:
                A type that is expected.
        """
        super().__init__("Too many None values.")
        self.keys = keys

    def __str__(self) -> Text:
        """A method for representing a text.

        A text value returning the expected value(s), which is/are unlike the
        expected value.

        @return: A Text.
        """
        return f"Expected one key(s) '{self.keys}' != None."


class WrongCurrencyPairFormatError(ValidationError):
    """
    Exception in case that a value has a wrong type.
    """

    def __init__(
            self,
            expected_value: List[str],
            actual_value: Union[Any],
            key: str):
        """
        Constructor of WrongCurrencyPairFormatError.

        Args:
            expected_value:
                A value that is expected.
            actual_value:
                The actual value, which is not the expected value.
            key:
                The incorrectly specified key.
        """
        super().__init__("Key has wrong value.")
        self.expected_value = expected_value
        self.actual_value = actual_value
        self.key = key

    def __str__(self) -> Text:
        """A method for representing a text.

        A text value returning the expected value(s), which is/are unlike the
        actual value.

        @return: A Text.
        """
        return f"Expected splitting value(s) '{self.expected_value}' != actual value '{self.actual_value}' " \
               f"in '{self.key}'."


class CustomBaseExceptionError(ValidationError):
    """
    Custom base exception
    """

    def __init__(self, key: Any, msg: Text):
        """
        Constructor of WrongCurrencyPairFormatError.

        @param key:
            The incorrectly specified key.
        @param msg:
            The exception message to return.
        """
        super().__init__("Custom base exception raised.")
        self.key = key
        self.msg = msg

    def __str__(self) -> Text:
        """A method for representing a text.

        A text value returning the expected value(s), which is/are unlike the
        actual value.

        @return: A Text.
        """
        return f"Error in key '{self.key}': {self.msg}."



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Singleton class providing a global configuration file.

Classes:
    - GlobalConfig
"""
from __future__ import annotations

import os
from typing import Optional


# ToDo Alle paths hier hinterlegen
class GlobalConfig(object):
    """
    Global config built after the Singleton pattern. That is, there can only exists ONE instance of a class or,
    equivalently, all instances share the same state.

    The config-file will be set and read only from utilities.read_config().
    """
    __instance: Optional[GlobalConfig] = None
    __is_initialized = False

    def __new__(cls) -> GlobalConfig:
        """
        Create a new instance of __GlobalConfig if and only if none exists. Else returns the existing instance.
        That ensures that all instances of the class share the same state.
        """
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)

        return cls.__instance

    def __init__(self) -> None:
        if not self.__is_initialized:
            self.__filename: Optional[str] = None
            self.__is_initialized = True

            self.path = os.getcwd() + "/resources/configs/user_configs/"
        try:
            # The first (os.getcwd()/...) is needed when the directory of the program and the resources differs.
            # That is the case for the Python Package installed via pip as we want the user to manipulate the resources
            # (i.e. config files and exchange mappings). The resources will be copied into the current working
            # directory and taken by the program from there.
            os.path.dirname(os.path.realpath(__file__)).index(self.path)
            self.path = os.path.dirname(os.path.realpath(__file__)) + "/user_configs/"
        except ValueError:
            pass

    def set_file(self, file: Optional[str] = None) -> None:
        """
        Sets self.__filename to either the variable given or an input string from the command prompt.
        The filename will be augmented with ".yaml" if needed.

        @param file: name of the config.
        @type file: str
        """
        if not file:
            file = input("Enter config file name [<filename> | quit/exit]: ").lower()
            if file in ["quit", "exit", "exit()", "quit()"]:
                raise SystemExit
                # sys.exit(0)
        if ".yaml" not in file:
            file = file + ".yaml"
        self.__filename = file

    @property  # TODO: Wrong use of property. Fix later with Steffen.
    def file(self, file: Optional[str] = None) -> str:
        """
        Returns the complete path to the config file. If the attribute self.__filename is not set yet,
        calls the method self.set_file.

        @param file: name of the config-file.
        @type file: str

        @return: complete path to config-file.
        @rtype: str
        """
        if not self.__filename:
            self.set_file(file=file)

        return "/".join([str(self.path), str(self.__filename)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# type: ignore[no-untyped-def]
"""
Test module for testing api_scheduling.Mapping class

Authors:
    Carolina Keil,
    Martin Schorfmann,
    Bjarne Schröder,
    Philipp Schönemann

Since:
    27.09.2018

Version:
    28.04.2021
"""

import datetime

import pytest

from model.exchange.mapping import Mapping


class TestMapping:
    """Test class for Mapping."""

    # pylint: disable=too-many-public-methods

    def test_extract_value_split_index_zero(self):
        """Test of splitting a str and taking the index zero."""

        mapping = Mapping("exchange_name",
                          ["product_id"],
                          ["str", "split", "-", 0])
        result = mapping.extract_value({
            "product_id": "BTC-ETH"
        })
        assert result == "BTC"

    def test_extract_value_split_index_one(self):
        """Test of splitting a str and taking the index one."""

        mapping = Mapping("exchange_name",
                          ["product_id"],
                          ["str", "split", "-", 1])
        result = mapping.extract_value({
            "product_id": "BTC-ETH"
        })
        assert result == "ETH"

    def test_extract_value_split_index_two(self):
        """Test of splitting a str and taking the index two."""

        mapping = Mapping("exchange_name",
                          ["product_id"],
                          ["str", "split", "-", 2])
        result = mapping.extract_value({
            "product_id": "BTC-ETH-USD"
        })
        assert result == "USD"

    def test_extract_value_slice_first_half(self):
        """Test of slicing a str and taking the first half."""

        mapping = Mapping("exchange_name",
                          ["product_id"],
                          ["str", "slice", 0, 3])
        result = mapping.extract_value({
            "product_id": "BTCETH"
        })
        assert result == "BTC"

    def test_extract_value_slice_second_half(self):
        """Test of slicing a str and taking the second half."""

        mapping = Mapping("exchange_name",
                          ["product_id"],
                          ["str", "slice", 3, 6])
        result = mapping.extract_value({
            "product_id": "BTCETH"
        })
        assert result == "ETH"

    # def test_extract_value_bool_to_str_true(self):
    #     """Test of conversion from bool to str in case of True."""
    #
    #     mapping = Mapping("active",
    #                       ["active"],
    #                       ["bool", "str"])
    #     result = mapping.extract_value({
    #         "active": True
    #     })
    #     self.assertIsInstance(result, str)
    #     self.assertEqual(result, "True")
    #
    # def test_extract_value_bool_to_str_false(self):
    #     """Test of conversion from bool to str in case of False."""
    #
    #     mapping = Mapping("active",
    #                       ["active"],
    #                       ["bool", "str"])
    #     result = mapping.extract_value({
    #         "active": False
    #     })
    #     self.assertIsInstance(result, str)
    #     self.assertEqual(result, "False")

    def test_extract_value_str_to_bool_true(self):
        """Test of conversion from str to bool in case of True."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "True"
        })

        assert isinstance(result, bool)
        assert result

    def test_extract_value_str_to_bool_true_lowercase(self):
        """Test of conversion from str to bool in case of lowercase True."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "true"
        })

        assert isinstance(result, bool)
        assert result

    def test_extract_value_str_to_bool_true_uppercase(self):
        """Test of conversion from str to bool in case of uppercase True."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "TRUE"
        })

        assert isinstance(result, bool)
        assert result

    def test_extract_value_str_to_bool_false(self):  #
        """Test of conversion from str to bool in case of False."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "False"
        })

        assert isinstance(result, bool)
        assert not result

    def test_extract_value_str_to_bool_false_lowercase(self):
        """Test of conversion from str to bool in case of lowercase False."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "false"
        })

        assert isinstance(result, bool)
        assert not result

    def test_extract_value_str_to_bool_false_uppercase(self):
        """Test of conversion from str to bool in case of uppercase False."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "FALSE"
        })

        assert isinstance(result, bool)
        assert not result

    def test_extract_value_str_to_bool_false_anything(self):
        """Test of conversion from str to bool in case of anything."""

        mapping = Mapping("active",
                          ["active"],
                          ["str", "bool"])

        result = mapping.extract_value({
            "active": "anything"
        })

        assert isinstance(result, bool)
        assert not result

    def test_extract_value_bool_to_int_true(self):
        """Test of conversion from bool to int in case of True."""

        mapping = Mapping("active",
                          ["active"],
                          ["bool", "int"])

        result = mapping.extract_value({
            "active": True
        })

        assert isinstance(result, int)
        assert result == 1

    def test_extract_value_bool_to_int_false(self):
        """Test of conversion from bool to int in case of False."""

        mapping = Mapping("active",
                          ["active"],
                          ["bool", "int"])

        result = mapping.extract_value({
            "active": False
        })

        assert isinstance(result, int)
        assert result == 0

    def test_extract_value_int_to_bool_one(self):
        """Test of conversion from int to bool in case of one."""

        mapping = Mapping("active",
                          ["active"],
                          ["int", "bool"])

        result = mapping.extract_value({
            "active": 1
        })

        assert isinstance(result, bool)
        assert result

    def test_extract_value_int_to_bool_zero(self):
        """Test of conversion from int to bool in case of zero."""

        mapping = Mapping("active",
                          ["active"],
                          ["int", "bool"])

        result = mapping.extract_value({
            "active": 0
        })

        assert isinstance(result, bool)
        assert not result

    def test_extract_value_int_to_bool_two(self):
        """Test of conversion from int to bool in case of two."""

        mapping = Mapping("active",
                          ["active"],
                          ["int", "bool"])

        result = mapping.extract_value({
            "active": 2
        })

        assert isinstance(result, bool)
        assert result

    def test_extract_value_int_to_bool_neg_two(self):
        """Test of conversion from int to bool in case of negative two."""

        mapping = Mapping("active",
                          ["active"],
                          ["int", "bool"])

        result = mapping.extract_value({
            "active": -2
        })

        assert isinstance(result, bool)
        assert result

    def test_extract_value_int_fromtimestamp(self):
        """Test of conversion from an int timestamp to datetime."""

        mapping = Mapping("time",
                          ["time"],
                          ["float", "from_timestamp", "0"])

        result = mapping.extract_value({
            "time": 1538122622
        })

        assert isinstance(result, datetime.datetime)
        # Different results depending on timezone
        # self.assertEqual(result, datetime.datetime(2018, 9, 28, 8, 17, 2))

    def test_extract_value_int_utcfromtimestamp(self):
        """Test of conversion from an int UTC timestamp to datetime."""

        mapping = Mapping("time",
                          ["time"],
                          ["float", "from_timestamp", "1"])

        result = mapping.extract_value({
            "time": 1538122622
        })

        assert isinstance(result, datetime.datetime)

    def test_extract_value_int_fromtimestampms(self):
        """Test of conversion from an int timestamp with ms to datetime."""

        mapping = Mapping("time",
                          ["time"],
                          ["float", "from_timestamp", "1"])

        result = mapping.extract_value({
            "time": 1538122622123
        })

        assert isinstance(result, datetime.datetime)
        assert result.microsecond == 123000

    def test_extract_value_float_fromtimestamp(self):
        """Test of conversion from a float timestamp to datetime."""

        mapping = Mapping("time",
                          ["time"],
                          ["float", "from_timestamp", "0"])

        result = mapping.extract_value({
            "time": 1538122622.123
        })

        assert isinstance(result, datetime.datetime)
        assert result.microsecond == 123000

    def test_extract_value_float_utcfromtimestamp(self):
        """Test of conversion from a float UTC timestamp to datetime."""

        mapping = Mapping("time",
                          ["time"],
                          ["float", "from_timestamp", "0"])

        result = mapping.extract_value({
            "time": 1538122622.123
        })

        assert isinstance(result, datetime.datetime)
        assert result.microsecond == 123000

    def test_extract_value_str_to_int_zero(self):
        """Test of conversion from str to int in case of zero."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "int"])

        result = mapping.extract_value({
            "number": "0"
        })

        assert isinstance(result, int)
        assert result == 0

    def test_extract_value_str_to_int_one(self):
        """Test of conversion from str to int in case of one."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "int"])

        result = mapping.extract_value({
            "number": "1"
        })

        assert isinstance(result, int)
        assert result == 1

    def test_extract_value_str_to_int_two(self):
        """Test of conversion from str to int in case of two."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "int"])

        result = mapping.extract_value({
            "number": "2"
        })

        assert isinstance(result, int)
        assert result == 2

    def test_extract_value_str_to_int_twelve(self):
        """Test of conversion from str to int in case of twelve."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "int"])

        result = mapping.extract_value({
            "number": "12"
        })

        assert isinstance(result, int)
        assert result == 12

    def test_extract_value_str_to_int_neg_one(self):
        """Test of conversion from str to int in case negative one."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "int"])

        result = mapping.extract_value({
            "number": "-1"
        })

        assert isinstance(result, int)
        assert result == -1

    def test_extract_value_str_to_float_zero(self):
        """Test of conversion from str to float in case of zero."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "0.0"
        })

        assert isinstance(result, float)
        assert result == 0.0

    def test_extract_value_str_to_float_one(self):
        """Test of conversion from str to float in case of one."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "1.0"
        })

        assert isinstance(result, float)
        assert result == 1.0

    def test_extract_value_str_to_float_two(self):
        """Test of conversion from str to float in case of two."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "2.0"
        })

        assert isinstance(result, float)
        assert result == 2.0

    def test_extract_value_str_to_float_twelve(self):
        """Test of conversion from str to float in case of twelve."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "12.0"
        })

        assert isinstance(result, float)
        assert result == 12.0

    def test_extract_value_str_to_float_pi(self):
        """Test of conversion from str to float in case of Pi."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "3.141592654"
        })

        assert isinstance(result, float)
        assert result == 3.141592654

    def test_extract_value_str_to_float_neg_one(self):
        """Test of conversion from str to float in case of negative one."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "-1.0"
        })

        assert isinstance(result, float)
        assert result == -1.0

    def test_extract_value_str_to_float_neg_pi(self):
        """Test of conversion from str to float in case of negative Pi."""

        mapping = Mapping("number",
                          ["number"],
                          ["str", "float"])

        result = mapping.extract_value({
            "number": "-3.141592654"
        })

        assert isinstance(result, float)
        assert result == -3.141592654

    def test_extract_value_datetime_totimestamp(self):
        """Test of conversion from datetome to timestamp."""

        mapping = Mapping("date",
                          ["date"],
                          ["datetime", "totimestamp"])

        result = mapping.extract_value({
            "date": datetime.datetime(2018, 10, 11, 11, 20)
        })

        assert isinstance(result, int)
        # Different results depending on timezone
        # self.assertEqual(result, 1539249600)

    def test_extract_value_datetime_totimestampms(self):
        """Test of conversion from datetome to timestamp with ms."""

        mapping = Mapping("date",
                          ["date"],
                          ["datetime", "totimestampms"])

        result = mapping.extract_value({
            "date": datetime.datetime(2018, 10, 11, 11, 20, 0, 123000)
        })

        assert isinstance(result, int)
        # Different results depending on timezone
        # self.assertEqual(result, 1539249600123)

    def test_extract_value_datetime_utctotimestamp(self):
        """Test of conversion from datetome to UTC timestamp."""

        mapping = Mapping("date",
                          ["date"],
                          ["datetime", "utctotimestamp"])

        result = mapping.extract_value({
            "date": datetime.datetime(2018, 10, 11, 11, 20, 0)
        })

        assert isinstance(result, int)
        # Different results depending on timezone

    def test_extract_value_str_strptime_date(self):
        """Test of conversion timestring via strptime in case of a date."""

        mapping = Mapping("date",
                          ["date"],
                          ["str", "strptime", "%Y-%m-%d"])

        result = mapping.extract_value({
            "date": "2018-10-11"
        })

        assert isinstance(result, datetime.datetime)
        assert result == datetime.datetime(2018, 10, 11)

    def test_extract_value_str_strptime_datetime(self):
        """Test of conversion timestring via strptime in case of a datetime."""

        mapping = Mapping("date",
                          ["date"],
                          ["str", "strptime", "%Y-%m-%d %H:%M"])

        result = mapping.extract_value({
            "date": "2018-10-11 12:06"
        })

        assert isinstance(result, datetime.datetime)
        assert result == datetime.datetime(2018, 10, 11, 12, 6)

    def test_extract_value_datetime_strftime_date(self):
        """Test of conversion from datetime via strftime in case of a date."""

        mapping = Mapping("date",
                          ["date"],
                          ["datetime", "strftime", "%Y-%m-%d"])

        result = mapping.extract_value({
            "date": datetime.datetime(2018, 10, 11)
        })

        assert isinstance(result, str)
        assert result == "2018-10-11"

    def test_extract_value_datetime_strftime_datetime(self):
        """Test of conversion from datetime via strftime in case of a dt."""

        mapping = Mapping("date",
                          ["date"],
                          ["datetime", "strftime", "%Y-%m-%d %H:%M"])

        result = mapping.extract_value({
            "date": datetime.datetime(2018, 10, 11, 12, 6)
        })

        assert isinstance(result, str)
        assert result == "2018-10-11 12:06"

    def test_extract_value_dict_key(self):
        """Test of extract value for dict_keys without further processing."""

        mapping = Mapping("pair",
                          ["dict_key"],
                          ["str"])

        result = mapping.extract_value({
            "BTC_USD": "value",
            "ETH_EUR": "value"
        })

        assert isinstance(result, list)
        assert result == ["BTC_USD", "ETH_EUR"]

    def test_extract_value_dict_key_split_index_zero(self):
        """Test of extract value for dict_keys with split and index 0."""

        mapping = Mapping("pair",
                          ["dict_key"],
                          ["str", "split", "_", 0])

        result = mapping.extract_value({
            "BTC_USD": "value",
            "ETH_EUR": "value"
        })

        assert isinstance(result, list)
        assert result == ["BTC", "ETH"]

    def test_extract_value_dict_key_split_index_one(self):
        """Test of extract value for dict_keys with split and index 1."""

        mapping = Mapping("pair",
                          ["dict_key"],
                          ["str", "split", "_", 1])

        result = mapping.extract_value({
            "BTC_USD": "value",
            "ETH_EUR": "value"
        })

        assert isinstance(result, list)
        assert result == ["USD", "EUR"]

    def test_extract_value_where_pair_can_not_be_split(self):
        """Test extract value for special case where value is currency pair but cannot be extracted
           since there is no delimiter for splitting the pair."""
        mapping = Mapping("first",
                          [],
                          ["first_currency"])
        result = mapping.extract_value(['ETHBTC', 'XRPBTC'], currency_pair_info=('XRP', 'BTC', 'XRPBTC'))
        assert result == "XRP"

        mapping = Mapping('second',
                          [],
                          ["second_currency"])
        result = mapping.extract_value(['ETHBTC', 'XRPBTC'], currency_pair_info=('ETH', 'BTC', 'ETHBTC'))
        assert result == "BTC"

    def test_extract_value_list(self):
        """Test of extracting all elements in a list."""
        mapping = Mapping("number_as_string",
                          [[]],
                          ['str'])
        value_list = ['1.0', '1.1', '1.2', '1.3', '2.4', '2.5', '3.6', '3.7']
        result = mapping.extract_value(value_list)

        assert value_list == result

    def test_extract_value_list_containing_dict_with_key(self):
        """Test of extract value for a list that contains a dict for each index with a known key."""

        # list element directly holds value behind key
        mapping = Mapping('first_currency',
                          ['first'],
                          ['str'])

        extract_list = [{'first': 'BTC'},
                        {'first': 'ETH'},
                        {'first': 'XRP'},
                        {'first': 'DOGE'},
                        {'first': 'XSA'},
                        {'first': 'TEST'}]
        result = mapping.extract_value(extract_list)
        value_list = ['BTC', 'ETH', 'XRP', 'DOGE', 'XSA', 'TEST']
        assert value_list == result

        # list element holds dict that holds another dict which holds the value behind key
        mapping = Mapping('first_currency',
                          ['a', 'first'],  # ['dict_values', 'first'] works also since it's only one level
                          ['str'])

        extract_list = [{'a': {'first': 'BTC', 'second': 'should not matter'}},
                        {'a': {'first': 'ETH', 'second': 'should not matter'}},
                        {'a': {'first': 'XRP', 'second': 'should not matter'}},
                        {'a': {'first': 'DOGE', 'second': 'should not matter'}},
                        {'a': {'first': 'XSA', 'second': 'should not matter'}},
                        {'a': {'first': 'TEST', 'second': 'should not matter'}}]
        result = mapping.extract_value(extract_list)
        assert value_list == result

        # list element with few more levels of dictionaries
        mapping = Mapping('first_currency',
                          ['a', 'b', 'c', 'd', 'first'],
                          ['str'])

        extract_list = [
            {'a': {'b': {'c': {'d': {'first': 'BTC', 'second': 'should not matter'}, 'dist': 'other_value'}}}},
            {'a': {'b': {'c': {'d': {'first': 'ETH', 'second': 'should not matter'}}, 'dist': 'other_value'}}},
            {'a': {'b': {'c': {'d': {'first': 'XRP', 'second': 'should not matter'}}}, 'dist': 'other_value'}},
            {'a': {'b': {'c': {'d': {'first': 'DOGE', 'second': 'should not matter'}}}}, 'dist': 'other_value'},
            {'a': {'b': {'c': {'d': {'first': 'XSA', 'second': 'should not matter'}}}, 'dist': 'other_value'}},
            {'a': {'b': {'c': {'d': {'first': 'TEST', 'second': 'should not matter'}}, 'dist': 'other_value'}}}]
        result = mapping.extract_value(extract_list)
        assert value_list == result

    def test_extract_values_dict_values(self):
        """Test of extract values for the special case that each key of a dictionary holds the values."""
        mapping = Mapping('first_currency',
                          ['dict_values'],
                          ['str'])

        extract_dict = {'1': 'XRP',
                        2: 'ETH',
                        '3': 'BTC',
                        4: 'DOGE'}
        result = mapping.extract_value(extract_dict)
        value_list = ['XRP', 'ETH', 'BTC', 'DOGE']
        assert value_list == result

        # extracting dicts with dict_values -> aka flattening the json
        mapping = Mapping('first_currency',
                          ['dict_values', 'first'],
                          ['str'])

        extract_dict = {'1': {'first': 'XRP', 'second': 'Whatever'},
                        2: {'first': 'ETH', 'second': 'Whatever'},
                        '3': {'first': 'BTC', 'second': 'Whatever'},
                        4: {'first': 'DOGE', 'second': 'Whatever'}}
        result = mapping.extract_value(extract_dict)
        value_list = ['XRP', 'ETH', 'BTC', 'DOGE']
        assert value_list == result

        # How to not use dict_values, since other values will be filtered also
        mapping = Mapping('first_currency',
                          ['dict_values', 'dict_values', 'dict_values', 'dict_values', 'first'],
                          ['str'])
        extract_list = [
            {'a': {'b': {'c': {'d': {'first': 'BTC', 'second': 'should not matter'}, 'dist': 'other_value'}}}},
            {'a': {'b': {'c': {'d': {'first': 'ETH', 'second': 'should not matter'}}, 'dist': 'other_value'}}},
            {'a': {'b': {'c': {'d': {'first': 'XRP', 'second': 'should not matter'}}}, 'dist': 'other_value'}},
            {'a': {'b': {'c': {'d': {'first': 'DOGE', 'second': 'should not matter'}}}}, 'dist': 'other_value'},
            {'a': {'b': {'c': {'d': {'first': 'XSA', 'second': 'should not matter'}}}, 'dist': 'other_value'}},
            {'a': {'b': {'c': {'d': {'first': 'TEST', 'second': 'should not matter'}}, 'dist': 'other_value'}}}]

        with pytest.raises(TypeError):
            mapping.extract_value(extract_list)

    def test_extract_value_list_containing_dict_where_key_is_value(self):
        """Test of extract value for a list that contains a dict that contains a dict.
           The first dict only has one key, which is the value we search for."""
        mapping = Mapping('second_currency',
                          ['dict_key'],
                          ['str', 'split', '_', 1])
        extract_list = [{'btc_eth': {'other': 'values'}},
                        {'btc_eur': {'other': 'values'}},
                        {'btc_usd': {'other': 'values'}},
                        {'btc_usdt': {'other': 'values'}}]
        value_list = ['eth', 'eur', 'usd', 'usdt']
        result = mapping.extract_value(extract_list)
        assert value_list == result

    def test_extract_value_list_containing_dict_where_pair_is_key_to_values(self):
        """Test of extract value for a list that contains dicts that hold a single dict.
           The key is the formatted currency pair that is needed to gain access to the value."""
        mapping = Mapping('value',
                          ['currency_pair', 1],
                          ['str', 'float'])
        extract_dict = {'btc_eth': {1: '123.456'},
                        'btc_xrp': {1: '789.101'},
                        'xrp_eth': {1: '112.131'},
                        'doge_btc': {1: '415.161'}}
        value_list = [415.161, 112.131, 789.101, 123.456]

        # note that currency pair infos are from bottom to top if compared to extract_dict
        currency_pair_infos = (('doge', 'BTC', 'doge_btc'),
                               ('xrp', 'eth', 'xrp_eth'),
                               ('btc', 'xrp', 'btc_xrp'),
                               ('btc', 'eth', 'btc_eth'))

        result = []
        for currency_pair_info in currency_pair_infos:
            result.append(mapping.extract_value(extract_dict, currency_pair_info=currency_pair_info))
        assert value_list == result

    def test_extract_value_dict_containing_list(self):
        """Test of extract value where the response is a dict and the known key contains a list of the values."""
        mapping = Mapping('first_currency',
                          ['data'],
                          ['str', 'split', '_', 0])
        extract_dict = {'data': ['eth_btc', 'usd_btc', 'xrp_eth', 'eth_xrp'],
                        'other_stuff': 'that does not matter'}
        value_list = ['eth', 'usd', 'xrp', 'eth']
        result = mapping.extract_value(extract_dict)
        assert value_list == result

    def test_extract_value_dict_containing_list_containing_dict_with_value(self):
        """Test of extract value where the response is a dict and a known key contains a list of dicts that hold the value."""
        mapping = Mapping('value',
                          ['data', 'value'],
                          ['str', 'float'])
        extract_dict = {'data': [{'value': '1.12'},
                                 {'value': '1.44'},
                                 {'value': '1.34'},
                                 {'value': '1.89'}]}
        value_list = [1.12, 1.44, 1.34, 1.89]
        result = mapping.extract_value(extract_dict)
        assert value_list == result

    def test_extract_value_dict_containing_list_containing_dict_where_key_is_value(self):
        """Test of extract value where the response is a dict and a known key contains a list of dicts with a single key that is the value."""
        mapping = Mapping('second_currency',
                          ['data', 'dict_key'],
                          ['str', 'split', '_', 1])
        extract_dict = {'data': [{'eth_btc': {'other': 'values'}},
                                 {'btc_xrp': {'other': 'values'}},
                                 {'tsla_usd': {'other': 'values'}},
                                 {'usdt_eth': {'other': 'values'}}]}
        value_list = ['btc', 'xrp', 'usd', 'eth']
        result = mapping.extract_value(extract_dict)
        assert value_list == result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module is preparing the python path to access all modules.
"""
import os
import sys
from pathlib import Path

sys.path.insert(0, os.path.dirname(__file__))

PATH_ABSOLUT = Path().parent.absolute()

all_paths = {
    "yaml_path": PATH_ABSOLUT.joinpath(Path("resources/running_exchanges/")),
    "program_config_path": PATH_ABSOLUT.joinpath(Path("resources/configs/program_config/config.yaml")),
    "path_absolut": Path().parent.absolute(),
    "template_path":  Path(os.path.dirname(os.path.realpath(__file__))).joinpath("resources/templates"),
    "user_config_path": PATH_ABSOLUT.joinpath("resources/configs/user_configs"),
    "package_path": os.path.dirname(os.path.realpath(__file__))
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Validator classes to validate the configuration file passed by the user. The module checks for existence, types and
values of necessary parameters. The file is executed in main.py before initializing the program.

Classes:
    - ConfigFileValidator(
        - ConfigYamlValidator(
            -> super().__init__(
            -    ConfigSectionValidator
            -    DatabaseStringValidator
            -    OperationSettingKeyValidator
            -    OperationSettingValueValidator
            -    RequestKeysValidator
            -    RequestValueValidator
            )
        )
    )
"""

from typing import Any, Text, Dict, Union, Optional

from pandas import Interval
from typeguard import check_type

# pylint: disable=too-many-lines
from model.validating.api_map_validators import LoadFileValidator, LoadYamlValidator
from model.validating.base import Report, Validator, CompositeValidator
from model.validating.errors import KeyNotInDictError, WrongTypeError, WrongValueError, WrongCompositeValueError
from model.validating.errors import WrongCurrencyPairFormatError, CustomBaseExceptionError


class ConfigFileValidator(CompositeValidator):
    """Validator for an API Map file.

    Validator for validating a given configuration file.
    Consists of a FileLoadValidator, a YamlLoadValidator and an ConfigYamlValidator.

    Attributes:
        value:
            The file name or file path as Text.
    """

    value: Text

    def validate(self) -> bool:
        """
        Validates the value.

        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """
        load_file = LoadFileValidator(self.value)
        is_file_loaded = load_file.validate()
        self.append_report(load_file)

        if not is_file_loaded:
            return False

        load_yaml = LoadYamlValidator(load_file)
        is_yaml_loaded = load_yaml.validate()
        self.append_report(load_yaml)

        if not is_yaml_loaded:
            return False

        config_file = ConfigYamlValidator(load_yaml.get_result_value())
        can_continue = config_file.validate()
        try:
            for report in config_file.report.reports:
                self.append_report(report)
        except (TypeError, Exception):
            self.append_report(config_file)

        return can_continue

    def result(self) -> bool:
        """
        TODO: Fill out
        """
        if not self.report:
            self.validate()

        return True


class ConfigYamlValidator(CompositeValidator):
    """
    Validator for loading Validators.

    Validator for calling the other required Validators.
    """

    def __init__(self, value: Dict[Text, Any]):
        """
        Constructor of ApiMapValidator.

        @param value: The value to get validated.
        """

        super().__init__(
            value,
            ConfigSectionValidator(value),
            DatabaseStringValidator(value["general"]["database"]),
            OperationSettingKeyValidator(value["general"]["operation_settings"]),
            OperationSettingValueValidator(value["general"]["operation_settings"]),
            RequestKeysValidator(value["jobs"]),
            RequestValueValidator(value["jobs"])
        )


class ConfigSectionValidator(Validator):
    """
    Validates if all sections and blocks are present.
    """
    block = ["general", "jobs"]
    section = ["database", "operation_settings"]

    def validate(self) -> bool:

        try:
            # ToDo: Falsch herum. So können nur fehlerhalfte Werte in der Config erkannt werden, jedoch nicht fehlende!
            for key in self.value:
                if key not in ConfigSectionValidator.block:
                    raise KeyNotInDictError(key, dict.fromkeys(ConfigSectionValidator.block))

            for key in self.value.get("general"):
                if key not in ConfigSectionValidator.section:
                    raise KeyNotInDictError(key, dict.fromkeys(ConfigSectionValidator.section))

        except KeyNotInDictError as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report(f"Configuration contains all blocks: {ConfigSectionValidator.block} and"
                                 f" sections: {ConfigSectionValidator.section}")
            return True


class DatabaseStringValidator(Validator):
    """
    Validates if all required database parameters exist to form the connection string.
    """
    db_strings = {"sqlite": ["sqltype", "db_name"],
                  "mariadb": ["sqltype", "client", "user_name", "password", "host", "port", "db_name"],
                  "mysql": ["sqltype", "client", "user_name", "password", "host", "port", "db_name"],
                  "postgresql": ["sqltype", "client", "user_name", "password", "host", "port", "db_name"],
                  }

    def validate(self) -> bool:
        """
        Validates the value.

        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        try:
            # check if sqltype exists in config file
            if "sqltype" not in self.value:
                raise KeyNotInDictError("sqltype", self.value)

            # Check if sqltype is correctly specified
            if self.value.get("sqltype") not in DatabaseStringValidator.db_strings.keys():
                raise KeyNotInDictError(self.value.get("sqltype"), DatabaseStringValidator.db_strings)

            # check if db-connection string is complete and not None
            for key in DatabaseStringValidator.db_strings.get(self.value.get("sqltype"), []):

                if key not in self.value:
                    raise KeyNotInDictError(key, self.value)

                if self.value.get(key) is None:
                    raise WrongTypeError([str, int], type(self.value.get(key)), key)

        except (KeyNotInDictError, WrongTypeError) as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Database connection string is valid")
            return True


class OperationSettingKeyValidator(Validator):
    """
    Validates if all necessary keys exist in the section 'operational_settings'.
    """

    sections = {"frequency": Union[str, int, float],
                "interval": Optional[str],
                "timeout": Union[int, float],
                }

    def validate(self) -> bool:
        """
        Validates the value.
        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        # Does key exist in config file
        try:
            for key, val in OperationSettingKeyValidator.sections.items():

                if key not in self.value:
                    raise KeyNotInDictError(key, dict.fromkeys(OperationSettingKeyValidator.sections))

                try:
                    check_type(key, self.value.get(key), val)
                except TypeError as error:
                    raise WrongTypeError(val, type(self.value.get(key))) from error

        except (KeyNotInDictError, WrongTypeError) as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Operation_settings have valid keys")
            return True


class OperationSettingValueValidator(Validator):
    """
    Validates if all necessary keys have correctly specified values in the section 'operational_settings'.
    """

    sections = {"frequency": Interval(0, 44640, "both"),  # max 31 days
                "interval": ["minutes", "hours", "days", "weeks", "months"],
                "timeout": Interval(0, 600, "both"),  # max 10 minutes
                "enable_logging": [True, False, 0, 1],
                "asynchronously": [True, False, 0, 1]
                }

    def validate(self) -> bool:
        """
        Validates the value.

        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        try:
            for key, val in OperationSettingValueValidator.sections.items():

                if key == "frequency" and isinstance(self.value.get(key), str):
                    if not self.value.get(key) == "once":
                        raise WrongValueError(["once", val], self.value.get(key), key)
                    continue

                if not self.value.get(key) in val:
                    raise WrongValueError(val, self.value.get(key), key)

        except WrongValueError as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Operation settings are valid")
            return True


class RequestKeysValidator(Validator):
    """
    Validates if all keys exist and types are correct for the request itself.
    """
    sections = {"request_method": str,
                "update_cp": Optional[bool],
                "exchanges": str,
                "excluded": Optional[str],
                "currency_pairs": Optional[str],
                "first_currencies": Optional[str],
                "second_currencies": Optional[str],
                }

    def validate(self) -> bool:
        """
        Validates the value.

        Validates the value attribute while generating a validation Report.

        @return: Whether further Validators may continue validating.
        """

        # if key is not nullable and not in the configuration file or empty
        try:
            for job in self.value:
                for key, val in RequestKeysValidator.sections.items():

                    if key not in self.value.get(job):
                        raise KeyNotInDictError(key, self.value.get(job))

                    try:
                        check_type(key, self.value.get(job).get(key), val)
                    except TypeError as error:
                        raise WrongTypeError(val, type(self.value.get(job).get(key)), key) from error

                    if self.value.get(job).get(key) == "None":
                        raise WrongValueError(["null"], self.value.get(job).get(key), key)

        except (KeyNotInDictError, WrongTypeError, WrongValueError) as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Request keys and types are valid.")
            return True


class RequestValueValidator(Validator):
    """
    Validates if exchange and currency-pairs are specified and not None.
    """
    VALID_JOBS = ["historic_rates", "tickers", "trades", "order_books", "currency_pairs"]

    def validate(self) -> bool:
        try:
            for job in self.value:

                if self.value.get(job).get("request_method") not in RequestValueValidator.VALID_JOBS:
                    raise WrongValueError(RequestValueValidator.VALID_JOBS, self.value.get(job).get("request_method"),
                                          "request_method")

                # if key 'currency-pairs' is specified
                if pair_string := self.value.get(job).get("currency_pairs"):

                    # Coinpaprika splits first currencies with "-", therefore "btc-bitcoin-usd" can't be read out.
                    if self.value.get(job).get("exchanges").lower() in ["coinpaprika"] and pair_string != 'all':
                        msg = "Specifying 'currency-pairs' is not allowed for Coinpaprika. " \
                              "Specify 'first-' and 'second-currencies' instead."
                        raise CustomBaseExceptionError("currency_pairs", msg)

                    # One way to check if the splitting value ("-") for currency-pairs and between (",") are
                    # correctly specified is to check if count("-") is >= 0 and always equal to count(",") + 1.
                    if pair_string.count("-") >= 0 and (pair_string.count("-") == pair_string.count(",") + 1):
                        continue
                    elif "all" in pair_string:
                        continue
                    else:
                        raise WrongCurrencyPairFormatError(["-", ","], pair_string, "currency_pairs")

                # if neither currency-pairs nor first_currencies or second_currencies are specified. That is only
                # allowed for the request method 'currency_pairs'.
                if self.value.get(job).get("request_method") == "currency_pairs":
                    continue
                if all([self.value.get(job).get("currency_pairs") is None,
                        self.value.get(job).get("first_currencies") is None,
                        self.value.get(job).get("second_currencies") is None]):
                    raise WrongCompositeValueError(["currency_pairs", "first_currencies", "second_currencies"])

        except (WrongCompositeValueError, KeyNotInDictError, WrongCurrencyPairFormatError,
                CustomBaseExceptionError, WrongValueError) as error:
            self.report = Report(error)
            return False

        else:
            self.report = Report("Currency-pairs are valid.")
            return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A command line tool to validate the exchange YAMLs.

How to use:
python validate.py { all | <exchange_name> }
"""

import os
import sys
from typing import Optional, Tuple, Union

import _paths
from model.validating.api_map_validators import ApiMapFileValidator
from model.validating.base import Report, CompositeReport
from model.validating.config_file_validator import ConfigFileValidator
from model.validating.program_config_validator import ProgramConfigValidator
from resources.configs.global_config import GlobalConfig

YAML_PATH = _paths.all_paths.get("yaml_path")
PATH = "open_crypto/resources/running_exchanges/"


def report_error(report: Report) -> Optional[Report]:
    """
    Recursive method to find the lowest False in a nested list
    """
    if report:
        return report

    if not isinstance(report, CompositeReport):
        return None

    for nested_report in report.reports:
        if not nested_report:
            try:
                report_error(nested_report)
            except AttributeError:
                print(nested_report)
                break


class ConfigValidator:
    """
    Class to validate the configuration file.
    """

    @staticmethod
    def validate_config_file() -> Tuple[bool, Union[Report, CompositeReport]]:
        """
        Calls the ConfigValidator
        :return: Validation result and report.
        """
        validator = ConfigFileValidator(GlobalConfig().file)
        if validator.validate():
            return True, validator.report
        else:
            return False, validator.report


class ProgramSettingValidator:
    """
    Class to validate the program configuration file.
    """

    @staticmethod
    def validate_config_file() -> Tuple[bool, Union[Report, CompositeReport]]:
        """
        Calls the ConfigValidator
        :return: Validation result and report.
        """
        validator = ProgramConfigValidator(_paths.all_paths.get("program_config_path"))
        if validator.validate():
            return True, validator.report
        else:
            return False, validator.report


class ExchangeValidator:
    """
    A validator that validates the YAML of a single exchange.
    """

    def __init__(self, exchange_name: str):
        """
        Create a new ExchangeValidator instance.

        @param exchange_name: The name of the exchange to be validated.
        """
        self.exchange_name = exchange_name

    def validate(self) -> bool:
        """
        Validate the exchange's YAML.

        @return: True if the YAML of the exchange is valid, False otherwise.
        """
        validator = ApiMapFileValidator(f"{PATH}/{self.exchange_name}.yaml")
        validator.validate()

        if validator.report:
            return True

        os.makedirs("reports/", exist_ok=True)
        with open("reports/report_" + self.exchange_name + ".txt", "w", encoding="UTF-8") as report:
            report.writelines(validator.report.indented_report())

        print("API Map is Invalid! \n"
              f"Inspect report: {'~/reports/report_' + self.exchange_name + '.txt'}")
        report_error(validator.report)
        return False


def validate_exchange(exchange_name: str) -> bool:
    """
    Validate the YAML of the specified exchange and print a human readable result.

    @param exchange_name: The exchange whose YAML is to be validated.

    @return: True if the YAML of the exchange is valid, False otherwise.
    """
    is_valid = ExchangeValidator(exchange_name).validate()
    print(f"Exchange: {exchange_name}, Valid: {is_valid}")
    return is_valid


if __name__ == "__main__":

    if len(sys.argv) == 1:
        print("How to use:")
        print("python validate.py { all | <exchange_name> }")
        sys.exit(1)

    exchange = sys.argv[1]

    if exchange != "all":
        sys.exit(int(not validate_exchange(exchange)))
    else:
        exchanges = [os.path.splitext(file)[0] for file in os.listdir(PATH) if file.endswith(".yaml")]

        valid_count = 0  # pylint: disable=C0103
        for exchange in exchanges:
            valid_count += int(validate_exchange(exchange))

        print(f"Valid Exchanges: {valid_count}/{len(exchanges)} ({int(valid_count / len(exchanges) * 100)} %)")
        sys.exit(int(valid_count != len(exchanges)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module initializes the jobs, starts the logger and finally the asyncio event-loop. Furthermore the
signal_handler adds the possibility to interrupt the process, the handler function to catch and log unexpected errors.
"""

import asyncio
import logging
import signal
import sys
from typing import Any, Dict, List

from model.database.db_handler import DatabaseHandler
from model.database.tables import metadata, ExchangeCurrencyPair
from model.exchange.exchange import Exchange
from model.scheduling.job import Job
from model.scheduling.scheduler import Scheduler
from model.utilities.kill_switch import KillSwitch
from model.utilities.loading_bar import Loader
from model.utilities.patch_event_loop import PatchEventLoop
from model.utilities.time_helper import TimeHelper
from model.utilities.utilities import read_config, yaml_loader, get_exchange_names, load_program_config
from model.utilities.utilities import signal_handler, init_logger, split_str_to_list, handler
from validate import ConfigValidator, ProgramSettingValidator

signal.signal(signal.SIGINT, signal_handler)


async def initialize_jobs(job_config: Dict[str, Any],
                          timeout: int,
                          interval: Any,
                          comparator: str,
                          db_handler: DatabaseHandler) -> List[Job]:
    """
    Initializes and creates new Job Objects and stores them in a list. There will be one Job-Object for every request
    method, independent of the amount of exchanges or currency_pairs specified in the config. The Dict
    'exchanges_with_pairs' is created with Exchange Objects as keys, the values are filled in the Scheduler.
    @param db_handler: Instance of the DatabaseHandler to pass to the Exchange Class. This is to be able to
                        perform database queries for variable request parameters.
    @param timeout: Request timeout for the Exchange Class.
    @param interval: Request Interval for HistoricRate (i.e. Daily, Minutes,..)
    @param comparator: Defines if the request interval of exchanges must match, be smaller or equal, ect.
    @param job_config: Dictionary with job parameter gathered from the config-file.
    @return: A list of Job objects.
    """
    jobs: List[Job] = list()

    for job in job_config.keys():
        job_params = job_config[job]

        if isinstance(job_params.get("exchanges"), str):
            job_params["exchanges"] = split_str_to_list(job_params.get("exchanges"))

        exchange_names = get_exchange_names() if "all" in job_params["exchanges"] else job_params["exchanges"]

        if job_params.get("excluded"):
            if isinstance(job_params.get("excluded"), str):
                job_params["exchanges"] = split_str_to_list(job_params.get("excluded"))
            exchange_names = [item for item in exchange_names if item not in job_params.get("excluded", [])]

        exchange_names = [yaml_loader(exchange) for exchange in exchange_names if yaml_loader(exchange) is not None]

        exchanges: [Exchange] = [Exchange(exchange_name,
                                          db_handler.get_first_timestamp,
                                          timeout,
                                          comparator=comparator,
                                          interval=interval) for exchange_name in exchange_names]

        exchanges_with_pairs: [Exchange, List[ExchangeCurrencyPair]] = dict.fromkeys(exchanges)

        new_job: Job = Job(job, job_params, exchanges_with_pairs)
        jobs.append(new_job)

    return jobs


async def main(database_handler: DatabaseHandler, program_config: dict) -> Scheduler.start:
    """
    The model() function to run the program. Loads the database, including the database_handler.
    The exchange_names are extracted with a helper method in utilities based on existing yaml-files.
    In an asynchronous manner it is iterated over the exchange and the responses are awaited and collected
        by await asyncio.gather(..)
    As soon as all responses from the exchange are returned, the values get extracted, formatted into tuples
        by the exchange.get_ticker(..) method and persisted by the into the database by the database_handler.

    @param: Instance of the DatabaseHandler
    @type database_handler: object
    @param program_config: Additional advanced program settings
    @type program_config: dict
    """

    with Loader("Initializing open_crypto...", ""):

        # ToDo: Disable exception hook automatically when terminating the program
        if program_config.get("exception_hook", True):
            sys.excepthook = handler

        job_config = read_config(file=None, section=None)
        operation_settings = job_config["general"]["operation_settings"]

        logging.info("Loading jobs.")
        jobs = await initialize_jobs(job_config=job_config["jobs"],
                                     timeout=operation_settings.get("timeout", 10),
                                     interval=operation_settings.get("interval", "days"),
                                     comparator=program_config["request_settings"].get("interval_settings",
                                                                                       "lower_or_equal"),
                                     db_handler=database_handler)

        frequency = operation_settings["frequency"]

    logging.info("Configuring Scheduler.")
    scheduler = Scheduler(database_handler, jobs, operation_settings.get("asynchronously", 1), frequency)
    await scheduler.validate_job()

    logging.info("Job(s) were created and will run with frequency: %s", frequency)

    while True:
        if not KillSwitch().stay_alive:
            print("Task got terminated.")
            logging.info("Task got terminated.")
            break

        if frequency == "once":
            loop = asyncio.get_event_loop()
            try:
                loop.run_until_complete(await scheduler.start())
            except (RuntimeError, TypeError) as exc:
                raise SystemExit from exc

        else:
            try:
                await scheduler.start()
            except Exception as ex:
                logging.exception(TimeHelper.now(), ex)


def run(file: str = None, path: str = None) -> None:
    """
    Starts the program and initializes the asyncio-event-loop.
    @param file: string representation of the configuration file
    @param path: String representation to the current working directory or any PATH specified in runner.py
    """

    program_config = load_program_config()

    db_params = read_config(file=file, section="database", reset=True)
    init_logger(path, program_config)

    logging.info("Validating user configuration files.")
    program_valid, program_report = ProgramSettingValidator.validate_config_file()
    config_valid, config_report = ConfigValidator.validate_config_file()
    validating_results = {program_valid: program_report, config_valid: config_report}

    for is_valid in validating_results.keys():
        if not is_valid:
            for nested_report in validating_results.get(is_valid).reports:
                print(nested_report)
                logging.error(nested_report)
            raise SystemExit

    logging.info("Establishing Database Connection")
    database_handler = DatabaseHandler(metadata,
                                       path=path,
                                       min_return_tuples=program_config["request_settings"].get("min_return_tuples", 2),
                                       **db_params)

    # ToDo: Still necessary?
    # See Github Issue for bug and work-around:
    # https://github.com/encode/httpx/issues/914
    if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform.startswith("win"):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    if PatchEventLoop.check_event_loop_exists():
        PatchEventLoop.apply_patch()

    asyncio.run(main(database_handler, program_config))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains possibilities to adjust several settings

Classes:
 - Setting: Contains advanced options for the program.
"""
import logging
import os
import shutil
from typing import Union, Any

import oyaml as yaml


class Settings:
    """
    Class to get and manipulate advanced program settings.
    """

    PATH = "/resources/configs/program_config/config.yaml"

    def __init__(self) -> None:
        self.config = Settings.get()
        self.copy = None

    @staticmethod
    def get() -> dict:
        """
        Returns the current program config.

        @return: The current program config.
        """
        with open(os.getcwd() + Settings.PATH, encoding="UTF-8") as file:
            return yaml.load(file, Loader=yaml.FullLoader)

    @staticmethod
    def _dump(config: dict) -> None:
        """
        Overwrites the current program config.

        @param config: The config to dump.
        """
        with open(os.getcwd() + Settings.PATH, "w", encoding="UTF-8") as file:
            yaml.dump(config, file)

    @staticmethod
    def set(block: str, key: str, val: Union[str, int, float]) -> None:
        """
        Sets a new value in the config file.

        @param block: Config file block
        @param key: Key within the block
        @param val: value to be set
        """
        try:
            with open(os.getcwd() + Settings.PATH, "r", encoding="UTF-8") as file:
                config = yaml.load(file, Loader=yaml.FullLoader)

            config.get(block).update({key: val})
            with open(os.getcwd() + Settings.PATH, "w", encoding="UTF-8") as file:
                yaml.dump(config, file)

        except (KeyError, FileNotFoundError):
            logging.error("Program config could not be updated.")
            print("Error updating program config!")

    @staticmethod
    def reset() -> None:
        """
        Creates a new template config file with all default settings.
        """
        filename = "program_config.yaml"

        source = os.path.dirname(os.path.realpath(__file__)) + "/resources/templates/"
        destination = os.getcwd() + "/resources/configs/program_config/"

        if os.path.exists(os.path.join(destination, filename)):
            os.remove(os.path.join(destination, filename.split("_")[1]))
        shutil.copy(os.path.join(source, filename),
                    os.path.join(destination, filename.split("_")[1]))

        print("Settings reset successful.")

    def _copy(self) -> None:
        """
        Copies the current config file.
        """
        self.copy = self.config.copy()

    def __enter__(self) -> object:
        """
        Enters the context manager.
        """
        self._copy()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """
        Exit the context manager.
        """
        self._dump(self.copy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class patches the open issue with nested asyncio EventLoops under several environments.
"""
import asyncio
import logging
from asyncio import AbstractEventLoop
from typing import Optional

import nest_asyncio


class PatchEventLoop:
    """
    Patch the nested-event-loop issue with several IDEs such as Jupyter-Notebook, Spyder, ect.
    """

    @staticmethod
    def _catch_event_loop() -> Optional[AbstractEventLoop]:
        """
        Catches an already running event-loop.

        @return: BaseEventLoop or None
        """
        return asyncio.get_event_loop()

    @staticmethod
    def check_event_loop_exists() -> bool:
        """
        Checks for an running event-loop.

        @return: bool.
        """
        return asyncio.get_event_loop().is_running()

    @staticmethod
    def apply_patch() -> None:
        """
        Prepares for nested event-loops or appends task to an already running event-loop.
        """
        logging.warning("You are attempting to start an nested asyncio EventLoop. \n"
                        "This is caused by design of several environments, such as Jupyter-Notebook or Spyder. \n"
                        "The current solution is the package 'nest_asyncio', which applies a patch. However, \n"
                        "if this package should depreciate in the future, extent this class by applying main.main() \n"
                        "as new task to the already running EventLoop.")
        nest_asyncio.apply()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module provides a kill switch to safely terminate the data collector.

Classes:
  - KillSwitch: Singleton Class.
"""

from __future__ import annotations

import threading
from typing import Optional, Any


class KillSwitch(object):
    """
    Global kill-switch built after the Singleton pattern. That is, there can only exists ONE instance of a class or,
    equivalently, all instances share the same state.

    The KillSwitch terminates the data collector after writing the data into the database and before making
    new requests.
    """
    __instance: Optional[KillSwitch] = None
    __is_initialized = False

    def __new__(cls) -> KillSwitch:
        """
        Creates a new instance of the class, if and only if, none exists.
        """
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)

        return cls.__instance

    def __init__(self) -> None:
        """
        Constructor
        """
        if not self.__is_initialized:
            self.__is_initialized = True
            self.stay_alive = True

    def kill(self) -> None:
        """
        Set state to False in order to kill process.
        """
        self.stay_alive = False

    def reset(self) -> None:
        """
        Reset the kill-switch.
        """
        self.stay_alive = True

    def set_timer(self, timer: int) -> None:
        """
        Sets the timer for the kill-switch.
        @param timer: Seconds to kill thread.
        """
        thread = threading.Timer(timer, self.kill)
        thread.start()

    def __enter__(self) -> object:
        """
        Enters the context manager.
        """
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> True:
        """
        Kill the process and exits the context manager.
        """
        self.reset()
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Class providing a loading bar animation. Either an instance of the class can be created or the class
may be used within the context manager framework.

Classes:
  - Loader
"""
import sys
from itertools import cycle
from shutil import get_terminal_size
from threading import Thread
from time import sleep
from typing import Any, Union

from colorama import Fore, Style, init

init()


class Loader:
    """
    Providing loading bar animation.

    Credit to: https://stackoverflow.com/a/66558182
    """

    def __init__(self, desc: str = "Loading...", end: str = "Done", timeout: float = 0.15, max_counter: int = None):
        """
        A loader-like context manager

        @param desc (str, optional): The loader's description. Defaults to "Loading...".
        @param end (str, optional): Final print. Defaults to "Done!".
        @param timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        @param max_counter: Maximum value of the counter.
        """
        self.desc = desc
        self.end = end
        self.timeout = timeout
        self.counter = 0
        self.max_count = max_counter if isinstance(max_counter, (int, float)) and max_counter > 1 else None

        self._thread = Thread(target=self._animate, daemon=True)
        self.steps = ["|", "/", "-", "\\"]
        self.done = False

    def start(self) -> object:
        """
        Starts the loading bar.
        @return self.
        """
        self._thread.start()
        return self

    def increment(self, step_size: Union[int, float] = 1) -> None:
        """
        Increments the counter.
        @param step_size: Incrementation size.
        """
        self.counter += step_size

    def _animate(self) -> None:
        """
        Prints the loading bar
        """
        for step in cycle(self.steps):
            if self.done:
                break
            if self.max_count:
                progress = f"{(self.counter / self.max_count) * 100:.2f}"
                print(f"\r{self.desc} {progress} % {step} ", flush=True, end="")
            else:
                print(f"\r{self.desc} {step}", flush=True, end="")
            sleep(self.timeout)

    def stop(self, color: str = "green", in_place: bool = False) -> None:
        """
        print text to console, a wrapper to built-in print

        @param color: can be one of "red" or "green", or "default"
        @param in_place: whether to erase previous line and print in place
        """

        # this ANSI code lets us erase the current line
        erase_line = "\x1b[2K"
        color_name_to_code = {"default": "", "red": Fore.RED, "green": Style.BRIGHT + Fore.GREEN}

        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        sys.stdout.write(color_name_to_code[color] + f"\r{self.end}" + Style.RESET_ALL)
        if in_place:
            print("\r" + erase_line, end="")

        sys.stdout.flush()

    def __enter__(self) -> object:
        """
        Enter class as context manager.
        """
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """
        Exits the context manager.
        """
        # handle exceptions with those variables
        self.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Setup module for installing the Crypto Data Collector including its dependencies.
"""

from setuptools import setup, find_packages

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="open-crypto",
    version="3.9.8",
    author="Steffen Guenther",
    author_email="crypto@uni-bremen.de",
    long_description=long_description,
    url="https://github.com/SteffenGue/open_crypto",
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        "aiohttp",
        "aioschedule",
        "certifi",
        "datetime_periods",
        "matplotlib",
        "numpy != 1.19.4",
        "oyaml",
        "pandas",
        "pytest",
        "python-dateutil",
        "sqlalchemy >= 1.4.22",
        "sqlalchemy_utils >= 0.37.8",
        "tqdm",
        "validators",
        "nest_asyncio",
        "typeguard >= 2.12.1",
        "colorama",
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Test module for requesting data
"""

import os
import asyncio
import pytest
import yaml

from model.database.tables import ExchangeCurrencyPair, Ticker, Currency, Trade
from model.database.tables import Exchange as DBExchange
from model.exchange.exchange import Exchange


class TestRequest:
    """
    Test class to perform several unit tests.
    """

    path = os.getcwd() + "/open_crypto/tests/unit_tests"

    with open(path + "/test_file.yaml", "r", encoding='UTF-8') as file:
        test_file: dict = yaml.load(file, Loader=yaml.FullLoader)
        # 'tickers' will be the non-existing request-method in this test file.
        test_file.get("requests").__delitem__("tickers")

    exchange = Exchange(test_file, None, None, None, None)

    def test_request_name_not_in_request_urls(self) -> None:
        """ Name of the request is not in the request_urls-dict."""

        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(TestRequest.exchange.request(Ticker, [], None))
        assert result is None

    def test_extract_request_urls_of_not_existing_request_method(self) -> None:
        """
        Test non existing request method
        """
        with pytest.raises(KeyError):
            TestRequest.exchange.extract_request_urls(TestRequest.test_file, 'tickers', Ticker, None)

    def test_apply_currency_pair_format(self) -> None:
        """
        Test apply currency-pair format specified in the exchange-yaml file.
        """
        exchange_object = DBExchange(name="TestExchange")
        first_currency = Currency(name="TestCurrency1")
        second_currency = Currency(name="TestCurrency2")
        exchange_currency_pair = ExchangeCurrencyPair(exchange=exchange_object, first=first_currency,
                                                      second=second_currency)
        request_urls = TestRequest.exchange.extract_request_urls(TestRequest.test_file['requests']['trades'], 'trades',
                                                                 Trade, exchange_currency_pair)
        TestRequest.exchange.request_urls = request_urls

        result = TestRequest.exchange.apply_currency_pair_format('trades', exchange_currency_pair)
        expected = 'TESTCURRENCY1-TESTCURRENCY2'
        assert result == expected

    def test_apply_currency_pair_format_with_missing_currency(self) -> None:
        """
        Test apply currency-pair format specified in the exchange-yaml file.
        """
        exchange_object = DBExchange(name="TestExchange")
        first_currency = Currency(name="TestCurrency1")

        exchange_currency_pair = ExchangeCurrencyPair(exchange=exchange_object, first=first_currency)
        request_urls = TestRequest.exchange.extract_request_urls(TestRequest.test_file['requests']['trades'], 'trades',
                                                                 Trade, exchange_currency_pair)
        TestRequest.exchange.request_urls = request_urls

        with pytest.raises(AttributeError):
            TestRequest.exchange.apply_currency_pair_format('trades', exchange_currency_pair)

#
# def test_request_name_empty_info(self):
#     """Name of the request is in request-urls dict, but has no value behind it."""
#     exchange_dict: Dict = {"name": 'test_exchange',
#                            "exchange": True,
#                            "api_url": 'https://url.to.api.com',
#                            "requests": []}
#     exchange = Exchange(exchange_dict, None, None, None, None)
#     exchange.request_urls = {"tickers": {}}
#     loop = asyncio.get_event_loop()
#     result = loop.run_until_complete(exchange.request(, []))
#     assert result is None

# def test_empty_currency_pairs(self):
#     """Given list of currency-pairs is empty or none."""
#     """Name of the request is in request-urls dict, but has no value behind it."""
#     exchange_dict: Dict = {"name": 'test_exchange',
#                            "exchange": True,
#                            "api_url": 'https://url.to.api.com',
#                            "requests": []}
#     exchange = Exchange(exchange_dict, None, None, None, None)
#     exchange.request_urls = {"tickers": {'url': "", "params": {}, "pair_template": {}}}
#     loop = asyncio.get_event_loop()
#     result = loop.run_until_complete(exchange.request(Ticker, [], None))
#     assert type(result[0]) == datetime
#     assert 'test_exchange' == result[1]
#     assert {} == result[2]

#     def test_failing_connection(self):
#         """Test where the session is unable to retreive information."""
#         http_mock = Mock(side_effect=ClientConnectionError)  # mock will raise error when get() is called
#         with patch('aiohttp.ClientSession.get', http_mock):
#             exchange_dict: Dict = {"name": 'test_exchange',
#                                    "exchange": True,
#                                    "api_url": 'https://url.to.api.com',
#                                    "requests": []}
#             exchange = Exchange(exchange_dict, None, None)
#             exchange.request_urls = {"tickers": {"url": {}, "params": {}, "pair_template": {}}}
#             loop = asyncio.get_event_loop()
#             ecp_mock = Mock(spec=ExchangeCurrencyPair)
#             result = loop.run_until_complete(exchange.request('tickers', [ecp_mock]))
#
#             http_mock.assert_called_with(url={}, params={})
#             assert type(result[0]) == datetime
#             assert 'test_exchange' == result[1]
#             assert {} == result[2]
#
#     def test_unreadable_response(self):
#         """Test where the response is not a json response."""
#         http_mock = Mock(return_value="error 404 response not found")
#         with patch('aiohttp.ClientSession.get', http_mock):
#             exchange_dict: Dict = {"name": 'test_exchange',
#                                    "exchange": True,
#                                    "api_url": 'https://url.to.api.com',
#                                    "requests": []}
#             exchange = Exchange(exchange_dict, None, None)
#             exchange.request_urls = {"tickers": {"url": {}, "params": {}, "pair_template": {}}}
#             ecp_mock = Mock(spec=ExchangeCurrencyPair)
#
#             loop = asyncio.get_event_loop()
#             result = loop.run_until_complete(exchange.request('tickers', [ecp_mock]))
#
#             http_mock.assert_called_with(url={}, params={})
#             assert type(result[0]) == datetime
#             assert 'test_exchange' == result[1]
#             assert {} == result[2]
#
#     def test_regular_request_for_all(self):
#         """Test for a request where all the information is retrieved with one request."""
#         json_mock = Mock(spec=ClientResponse)
#         json_mock.json.return_value = {"data": 'This is the response.'}
#         http_mock = AsyncMock(return_value=json_mock)
#         with patch('aiohttp.ClientSession.get', http_mock):
#             exchange_dict: Dict = {"name": 'test_exchange',
#                                    "exchange": True,
#                                    "api_url": 'https://url.to.api.com',
#                                    "requests": []}
#             exchange = Exchange(exchange_dict, None, None)
#             exchange.request_urls = {"tickers": {"url": {}, "params": {}, "pair_template": {}}}
#             ecp_mock = Mock(spec=ExchangeCurrencyPair)
#
#             loop = asyncio.get_event_loop()
#             result = loop.run_until_complete(exchange.request('tickers', [ecp_mock]))
#             json_mock.json.assert_called_once()
#             http_mock.assert_called_with(url={}, params={})
#             assert type(result[0]) == datetime
#             assert 'test_exchange' == result[1]
#             assert {None: {'data': 'This is the response.'}} == result[2]
#
#     def test_request_with_params(self):
#         """Test for a regular request where parameters have to be sent."""
#         json_mock = Mock(spec=ClientResponse)
#         json_mock.json.return_value = {"data": 'This is the response.'}
#         http_mock = AsyncMock(return_value=json_mock)
#         with patch('aiohttp.ClientSession.get', http_mock):
#             exchange_dict: Dict = {"name": 'test_exchange',
#                                    "exchange": True,
#                                    "api_url": 'https://url.to.api.com',
#                                    "requests": []}
#             exchange = Exchange(exchange_dict)
#             exchange.request_urls = {"tickers": {"url": {},
#                                                  "params": {'steps': 1299, 'query_all': True},
#                                                  "pair_template": {}}}
#             ecp_mock = Mock(spec=ExchangeCurrencyPair)
#
#             loop = asyncio.get_event_loop()
#             result = loop.run_until_complete(exchange.request('tickers', [ecp_mock]))
#             json_mock.json.assert_called_once()
#             http_mock.assert_called_with(url={}, params={'steps': 1299, 'query_all': True})
#             assert type(result[0]) == datetime
#             assert 'test_exchange' == result[1]
#             assert {None: {'data': 'This is the response.'}} == result[2]
#
#     def test_request_with_pair_formatting_in_url(self):
#         """Test for a request where the pair info has to be sent as a string in a certain format."""
#         json_mock = Mock(spec=ClientResponse)
#         json_mock.json.return_value = {"data": 'This is the response.'}
#         http_mock = AsyncMock(return_value=json_mock)
#         with patch('aiohttp.ClientSession.get', http_mock):
#             exchange_dict: Dict = {"name": 'test_exchange',
#                                    "exchange": True,
#                                    "api_url": 'https://url.to.api.com',
#                                    "requests": []}
#             exchange = Exchange(exchange_dict, None, None)
#             exchange.request_urls = {"tickers": {"url": 'tickers/{currency_pair}',
#                                                  "params": {},
#                                                  "pair_template": {'template': '{first}_{second}', 'lower_case': True}}}
#             ecp_mock = Mock(spec=ExchangeCurrencyPair)
#             ecp_mock.first.name = 'BTC'
#             ecp_mock.second.name = 'ETH'
#
#             loop = asyncio.get_event_loop()
#             result = loop.run_until_complete(exchange.request('tickers', [ecp_mock]))
#             json_mock.json.assert_called_once()
#             http_mock.assert_called_with(url='tickers/btc_eth', params={})
#             assert type(result[0]) == datetime
#             assert 'test_exchange' == result[1]
#             assert {None: {'data': 'This is the response.'}} == result[2]
#
#     def test_request_with_pair_formatting_as_param(self):
#         """Test for a request where the pair info has to be sent as a formatted string as parameter."""
#         json_mock = Mock(spec=ClientResponse)
#         json_mock.json.return_value = {"data": 'This is the response.'}
#         http_mock = AsyncMock(return_value=json_mock)
#         with patch('aiohttp.ClientSession.get', http_mock):
#             exchange_dict: Dict = {"name": 'test_exchange',
#                                    "exchange": True,
#                                    "api_url": 'https://url.to.api.com',
#                                    "requests": []}
#             exchange = Exchange(exchange_dict, None, None)
#             exchange.request_urls = {"tickers": {"url": '',
#                                                  "params": {},
#                                                  "pair_template": {'template': '{first}_{second}', 'lower_case': True,
#                                                                    'alias': 'pair'}}}
#             ecp_mock = Mock(spec=ExchangeCurrencyPair)
#             ecp_mock.first.name = 'BTC'
#             ecp_mock.second.name = 'ETH'
#
#             loop = asyncio.get_event_loop()
#             result = loop.run_until_complete(exchange.request('tickers', [ecp_mock]))
#             json_mock.json.assert_called_once()
#             http_mock.assert_called_with(url='', params={'pair': 'btc_eth'})
#             assert type(result[0]) == datetime
#             assert 'test_exchange' == result[1]
#             assert {None: {'data': 'This is the response.'}} == result[2]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module providing a job which defines the request method and exchanges currency-pairs.

Classes:
    - Job
"""
from typing import Any, Dict, Optional

from model.database.tables import ExchangeCurrencyPair
from model.exchange.exchange import Exchange


class Job:
    """
    Represents a job that is executed by the Scheduler based on the frequency set in config.yaml.
    A job itself is 'created' in config.yaml.
    Should only be used to hold information for it tasks.
    """

    def __init__(self,
                 name: str,
                 job_params: Dict[str, Any],
                 exchanges_with_pairs: Dict[Exchange, Dict[ExchangeCurrencyPair, Optional[int]]]):
        """
        Initializer of a job.

        @param name: Name of the job taken out of config-file
        @type name: str
        @param job_params: All job parameter from the config-file
        @type job_params: dict
        @param exchanges_with_pairs: Dict for each exchange and it's currency pairs that have to be queried from it
                                     together with the last inserted row_id, if available..
        @type exchanges_with_pairs: dict[Exchange, Dict[ExchangeCurrencyPair], Optional[int]]
        """
        self.name = name
        self.request_name = job_params["request_method"]
        self.exchanges_with_pairs = exchanges_with_pairs
        self.job_params = job_params
